from __future__ import annotations

from typing import List

from AoE2ScenarioParser.datasets.support.info_dataset_base import InfoDatasetBase


class UnitInfo(InfoDatasetBase):
    """
    **Description**

    This class provides information about most of the units in the game. Information about the following properties of
    a unit is found in this class:
     - Unit ID
     - Icon ID
     - Dead Unit ID
     - HotKey ID
     - If the unit is a gaia only unit (eg. deer, boar, etc.)

    **Methods**

    >>> UnitInfo.vils()
    >>> UnitInfo.unique_units()

    **Inherited Methods from class InfoDatasetBase**

    >>> InfoDatasetBase.from_id()
    >>> InfoDatasetBase.from_dead_id()
    >>> InfoDatasetBase.from_icon_id()
    >>> InfoDatasetBase.from_hotkey_id()
    >>> InfoDatasetBase.gaia_only()
    >>> InfoDatasetBase.non_gaia()

    **Examples**

    >>> UnitInfo.VILLAGER_FEMALE.ID
    >>> 293

    >>> UnitInfo.PETARD.ICON_ID
    >>> 113

    >>> UnitInfo.PALADIN.DEAD_ID
    >>> 570

    >>> UnitInfo.VILLAGER_MALE.HOTKEY_ID
    >>> 16121

    >>> UnitInfo.DEER.IS_GAIA_ONLY
    >>> True
    """

    @staticmethod
    def vils(exclude_female: bool = False, exclude_male: bool = False) -> List[UnitInfo]:
        """
        Args:
            exclude_female: if set to true, exclude the female villagers
            exclude_male: if set to true, exclude the male villagers

        Returns:
            A list of villager UnitInfo objects
        """
        villagers = {
            "male": [
                UnitInfo.VILLAGER_MALE,
                UnitInfo.VILLAGER_MALE_BUILDER,
                UnitInfo.VILLAGER_MALE_FARMER,
                UnitInfo.VILLAGER_MALE_FISHERMAN,
                UnitInfo.VILLAGER_MALE_FORAGER,
                UnitInfo.VILLAGER_MALE_GOLD_MINER,
                UnitInfo.VILLAGER_MALE_HUNTER,
                UnitInfo.VILLAGER_MALE_LUMBERJACK,
                UnitInfo.VILLAGER_MALE_REPAIRER,
                UnitInfo.VILLAGER_MALE_SHEPHERD,
                UnitInfo.VILLAGER_MALE_STONE_MINER,
            ],
            "female": [
                UnitInfo.VILLAGER_FEMALE,
                UnitInfo.VILLAGER_FEMALE_BUILDER,
                UnitInfo.VILLAGER_FEMALE_FARMER,
                UnitInfo.VILLAGER_FEMALE_FISHERMAN,
                UnitInfo.VILLAGER_FEMALE_FORAGER,
                UnitInfo.VILLAGER_FEMALE_GOLD_MINER,
                UnitInfo.VILLAGER_FEMALE_HUNTER,
                UnitInfo.VILLAGER_FEMALE_LUMBERJACK,
                UnitInfo.VILLAGER_FEMALE_REPAIRER,
                UnitInfo.VILLAGER_FEMALE_SHEPHERD,
                UnitInfo.VILLAGER_FEMALE_STONE_MINER,
            ]
        }

        units_to_return = []

        if not exclude_female:
            units_to_return.extend(villagers["female"])
        if not exclude_male:
            units_to_return.extend(villagers["male"])

        return units_to_return

    @staticmethod
    def unique_units(
            exclude_elite_units: bool = False,
            exclude_non_elite_units: bool = False,
            exclude_castle_units: bool = False,
            exclude_non_castle_units: bool = False
    ) -> List[UnitInfo]:
        """
        Args:
            exclude_elite_units: if set to false, exclude the elite unique units
            exclude_non_elite_units: if set to false, exclude the non elite unique units
            exclude_castle_units: if set to false, exclude the castle unique units
            exclude_non_castle_units: if set to false, excludes the unique units not trained at the castle

        Returns:
            A list of unique unit UniInfo objects
        """
        unique_units = {
            "castle": {
                "non_elite": [
                    UnitInfo.ARAMBAI,
                    UnitInfo.BALLISTA_ELEPHANT,
                    UnitInfo.BERSERK,
                    UnitInfo.BOYAR,
                    UnitInfo.CAMEL_ARCHER,
                    UnitInfo.CAMEL_SCOUT,
                    UnitInfo.CATAPHRACT,
                    UnitInfo.CHAKRAM_THROWER,
                    UnitInfo.CHU_KO_NU,
                    UnitInfo.CONQUISTADOR,
                    UnitInfo.COUSTILLIER,
                    UnitInfo.FLAMING_CAMEL,
                    UnitInfo.GBETO,
                    UnitInfo.GENOESE_CROSSBOWMAN,
                    UnitInfo.GHULAM,
                    UnitInfo.HUSKARL,
                    UnitInfo.HUSSITE_WAGON,
                    UnitInfo.JAGUAR_WARRIOR,
                    UnitInfo.JANISSARY,
                    UnitInfo.KAMAYUK,
                    UnitInfo.KARAMBIT_WARRIOR,
                    UnitInfo.KESHIK,
                    UnitInfo.KIPCHAK,
                    UnitInfo.KONNIK,
                    UnitInfo.LEITIS,
                    UnitInfo.LONGBOWMAN,
                    UnitInfo.MAGYAR_HUSZAR,
                    UnitInfo.MAMELUKE,
                    UnitInfo.MANGUDAI,
                    UnitInfo.OBUCH,
                    UnitInfo.ORGAN_GUN,
                    UnitInfo.PLUMED_ARCHER,
                    UnitInfo.RATHA_MELEE,
                    UnitInfo.RATHA_RANGED,
                    UnitInfo.RATTAN_ARCHER,
                    UnitInfo.SAMURAI,
                    UnitInfo.SERJEANT,
                    UnitInfo.SHOTEL_WARRIOR,
                    UnitInfo.TARKAN,
                    UnitInfo.TEUTONIC_KNIGHT,
                    UnitInfo.THIRISADAI,
                    UnitInfo.THROWING_AXEMAN,
                    UnitInfo.URUMI_SWORDSMAN,
                    UnitInfo.WAR_ELEPHANT,
                    UnitInfo.WAR_WAGON,
                    UnitInfo.WOAD_RAIDER,
                ],
                "elite": [
                    UnitInfo.ELITE_ARAMBAI,
                    UnitInfo.ELITE_BALLISTA_ELEPHANT,
                    UnitInfo.ELITE_BERSERK,
                    UnitInfo.ELITE_BOYAR,
                    UnitInfo.ELITE_CAMEL_ARCHER,
                    UnitInfo.ELITE_CATAPHRACT,
                    UnitInfo.ELITE_CHAKRAM_THROWER,
                    UnitInfo.ELITE_CHU_KO_NU,
                    UnitInfo.ELITE_CONQUISTADOR,
                    UnitInfo.ELITE_COUSTILLIER,
                    UnitInfo.ELITE_GBETO,
                    UnitInfo.ELITE_GENOESE_CROSSBOWMAN,
                    UnitInfo.ELITE_GHULAM,
                    UnitInfo.ELITE_HUSKARL,
                    UnitInfo.ELITE_HUSSITE_WAGON,
                    UnitInfo.ELITE_JAGUAR_WARRIOR,
                    UnitInfo.ELITE_JANISSARY,
                    UnitInfo.ELITE_KAMAYUK,
                    UnitInfo.ELITE_KARAMBIT_WARRIOR,
                    UnitInfo.ELITE_KESHIK,
                    UnitInfo.ELITE_KIPCHAK,
                    UnitInfo.ELITE_KONNIK,
                    UnitInfo.ELITE_LEITIS,
                    UnitInfo.ELITE_LONGBOWMAN,
                    UnitInfo.ELITE_MAGYAR_HUSZAR,
                    UnitInfo.ELITE_MAMELUKE,
                    UnitInfo.ELITE_MANGUDAI,
                    UnitInfo.ELITE_OBUCH,
                    UnitInfo.ELITE_ORGAN_GUN,
                    UnitInfo.ELITE_PLUMED_ARCHER,
                    UnitInfo.ELITE_RATHA_MELEE,
                    UnitInfo.ELITE_RATHA_RANGED,
                    UnitInfo.ELITE_RATTAN_ARCHER,
                    UnitInfo.ELITE_SAMURAI,
                    UnitInfo.ELITE_SERJEANT,
                    UnitInfo.ELITE_SHOTEL_WARRIOR,
                    UnitInfo.ELITE_TARKAN,
                    UnitInfo.ELITE_TEUTONIC_KNIGHT,
                    UnitInfo.ELITE_THROWING_AXEMAN,
                    UnitInfo.ELITE_URUMI_SWORDSMAN,
                    UnitInfo.ELITE_WAR_ELEPHANT,
                    UnitInfo.ELITE_WAR_WAGON,
                    UnitInfo.ELITE_WOAD_RAIDER,
                ]
            },
            "non_elite": [
                UnitInfo.CARAVEL,
                UnitInfo.CONDOTTIERO,
                UnitInfo.CONDOTTIERO_PLACEHOLDER,
                UnitInfo.FLEMISH_MILITIA,
                UnitInfo.FLEMISH_MILITIA_FEMALE,
                UnitInfo.FLEMISH_MILITIA_MALE,
                UnitInfo.GENITOUR,
                UnitInfo.HOUFNICE,
                UnitInfo.HUSKARL_BARRACKS,
                UnitInfo.IMPERIAL_CAMEL_RIDER,
                UnitInfo.IMPERIAL_SKIRMISHER,
                UnitInfo.KONNIK_DISMOUNTED,
                UnitInfo.KONNIK_KREPOST,
                UnitInfo.LONGBOAT,
                UnitInfo.MISSIONARY,
                UnitInfo.SERJEANT_DONJON,
                UnitInfo.SHRIVAMSHA_RIDER,
                UnitInfo.SLINGER,
                UnitInfo.TARKAN_STABLE,
                UnitInfo.TURTLE_SHIP,
                UnitInfo.WINGED_HUSSAR
            ],
            "elite": [
                UnitInfo.ELITE_CARAVEL,
                UnitInfo.ELITE_GENITOUR,
                UnitInfo.ELITE_HUSKARL_BARRACKS,
                UnitInfo.ELITE_KIPCHAK_CUMAN_MERCENARIES,
                UnitInfo.ELITE_KONNIK_DISMOUNTED,
                UnitInfo.ELITE_KONNIK_KREPOST,
                UnitInfo.ELITE_LONGBOAT,
                UnitInfo.ELITE_SERJEANT_DONJON,
                UnitInfo.ELITE_SHRIVAMSHA_RIDER,
                UnitInfo.ELITE_TARKAN_STABLE,
                UnitInfo.ELITE_TURTLE_SHIP,
            ]
        }

        units_to_return = []

        if not exclude_non_elite_units:
            if not exclude_non_castle_units:
                units_to_return.extend(unique_units["non_elite"])
            if not exclude_castle_units:
                units_to_return.extend(unique_units["castle"]["non_elite"])
        if not exclude_elite_units:
            if not exclude_non_castle_units:
                units_to_return.extend(unique_units["elite"])
            if not exclude_castle_units:
                units_to_return.extend(unique_units["castle"]["elite"])

        return units_to_return

    BEAR = 486, 151, 489, 16712, True
    BUTTERFLY1 = 1608, -1, -1, 16716, True
    BUTTERFLY2 = 1609, -1, -1, 16716, True
    BUTTERFLY3 = 1610, -1, -1, 16716, True
    CROCODILE = 1031, 193, 1032, 16060, True
    DEER = 65, 3, 43, 16071, True
    DIRE_WOLF = 89, 7, 237, 16553, True
    ELEPHANT = 1301, 179, 1332, 16722, True
    FALCON = 1056, 5, -1, 16058, True
    HAWK = 96, 283, -1, 16073, True
    IBEX = 1239, 267, 1240, 16071, True
    IRON_BOAR = 810, 98, 356, 16710, True
    JAGUAR = 812, 111, 813, 16712, True
    JAVELINA = 822, 98, 356, 16722, True
    KOMODO_DRAGON = 1135, 226, 1136, 16163, True
    LION = 1029, 194, 1030, 16059, True
    MACAW = 816, 284, -1, 16716, True
    OSTRICH = 1026, 196, 1027, 16071, True
    RABID_WOLF = 202, 7, 237, 16652, True
    RHINOCEROS = 1139, 225, 1140, 16172, True
    SEAGULLS = 303, -1, -1, 16716, True
    SNOW_LEOPARD = 1241, 268, 1242, 16060, True
    STORK = 1028, 285, -1, 16058, True
    STORMY_DOG = 862, 159, -1, 16764, True
    TIGER = 1137, 227, 1138, 16170, True
    VULTURE = 1305, 286, -1, 16716, True
    WILD_BACTRIAN_CAMEL = 1247, 266, 1238, 16735, True
    WILD_BOAR = 48, 98, 356, 16406, True
    WILD_CAMEL = 884, 135, 898, 16735, True
    WILD_HORSE = 835, 112, 815, 16735, True
    WOLF = 126, 7, 237, 16075, True
    ZEBRA = 1019, 192, 1020, 16071, True
    MOVEABLE_MAP_REVEALER = 0, -1, -1, 16000, False
    ALFRED_THE_ALPACA = 1300, 177, 1331, 16469, False
    AMAZON_ARCHER = 850, 165, 1325, 16683, False
    AMAZON_WARRIOR = 825, 166, 1324, 16667, False
    ARAMBAI = 1126, 230, 1127, 16101, False
    ARBALESTER = 492, 90, 496, 16418, False
    ARCHER = 4, 17, 3, 16083, False
    BACTRIAN_CAMEL = 1237, 266, 1238, 16714, False
    BALLISTA_ELEPHANT = 1120, 231, 1121, 16101, False
    BANDIT = 299, 8, 152, 16098, False
    BATTERING_RAM = 1258, 74, 23, 16094, False
    BATTLE_ELEPHANT = 1132, 228, 1133, 16733, False
    BERSERK = 692, 38, 693, 16574, False
    BOMBARD_CANNON = 36, 30, 16, 16093, False
    BOYAR = 876, 114, 877, 16101, False
    CAMEL = 897, 135, 898, 16714, False
    CAMEL_ARCHER = 1007, 191, 1008, 16101, False
    CAMEL_RIDER = 329, 78, 113, 16416, False
    CANNON_GALLEON = 420, 55, -1, 16287, False
    CANOE = 778, 164, -1, 16106, False
    CAPPED_RAM = 422, 63, 423, 16289, False
    CARAVEL = 1004, 198, -1, 16106, False
    CART = 1338, 34, 205, 16100, False
    CATAPHRACT = 40, 35, 27, 16101, False
    CAVALIER = 283, 49, 139, 16070, False
    CAVALRY_ARCHER = 39, 19, 34, 16085, False
    CENTURION = 275, 138, 277, 16451, False
    CHAMPION = 567, 72, 568, 16469, False
    CHU_KO_NU = 73, 36, 28, 16102, False
    COBRA_CAR = 748, 142, -1, 16658, False
    CONDOTTIERO = 882, 134, 883, 16679, False
    CONQUISTADOR = 771, 106, 772, 16687, False
    COW_A = 705, 150, 843, 16498, False
    COW_B = 1596, 150, 1597, 16498, False
    COW_C = 1598, 150, 1599, 16498, False
    COW_D = 1600, 150, 1601, 16498, False
    CROSSBOWMAN = 24, 18, 26, 16084, False
    DEMOLITION_RAFT = 1104, 202, -1, 16424, False
    DEMOLITION_SHIP = 527, 84, -1, 16424, False
    DONKEY = 846, 158, 848, 16100, False
    DRAGON_SHIP = 1302, 178, -1, 16739, False
    EAGLE_SCOUT = 751, 109, 754, 16671, False
    EAGLE_WARRIOR = 753, 148, 1116, 16671, False
    EASTERN_SWORDSMAN = 894, 186, 895, 16081, False
    ELEPHANT_ARCHER = 873, 93, 874, 16101, False
    ELITE_ARAMBAI = 1128, 230, 1127, 16101, False
    ELITE_BALLISTA_ELEPHANT = 1122, 231, 1121, 16101, False
    ELITE_BATTLE_ELEPHANT = 1134, 246, 1154, 16733, False
    ELITE_BERSERK = 694, 38, 695, 16576, False
    ELITE_BOYAR = 878, 114, 877, 16101, False
    ELITE_CAMEL_ARCHER = 1009, 191, 1008, 16101, False
    ELITE_CANNON_GALLEON = 691, 298, -1, 16573, False
    ELITE_CARAVEL = 1006, 198, -1, 16106, False
    ELITE_CATAPHRACT = 553, 35, 27, 16451, False
    ELITE_CHU_KO_NU = 559, 36, 28, 16452, False
    ELITE_CONQUISTADOR = 773, 106, 772, 16689, False
    ELITE_EAGLE_WARRIOR = 752, 149, 1117, 16673, False
    ELITE_ELEPHANT_ARCHER = 875, 93, 874, 16101, False
    ELITE_GBETO = 1015, 197, 1014, 16101, False
    ELITE_GENITOUR = 1012, 201, 1011, 16417, False
    ELITE_GENOESE_CROSSBOWMAN = 868, 133, 867, 16101, False
    ELITE_HUSKARL = 555, 50, 62, 16454, False
    ELITE_JAGUAR_WARRIOR = 726, 110, 750, 16669, False
    ELITE_JANISSARY = 557, 39, 107, 16455, False
    ELITE_KAMAYUK = 881, 97, 880, 16460, False
    ELITE_KARAMBIT_WARRIOR = 1125, 233, 1124, 16101, False
    ELITE_KESHIK = 1230, 251, 1229, 16101, False
    ELITE_KIPCHAK = 1233, 252, 1232, 16108, False
    ELITE_KONNIK = 1227, 249, 1253, 16101, False
    ELITE_KONNIK_DISMOUNTED = 1253, 250, 1257, 16411, False
    ELITE_LEITIS = 1236, 253, 1235, 16101, False
    ELITE_LONGBOAT = 533, 40, -1, 16457, False
    ELITE_LONGBOWMAN = 530, 41, 115, 16456, False
    ELITE_MAGYAR_HUSZAR = 871, 99, 870, 16101, False
    ELITE_MAMELUKE = 556, 37, 44, 16453, False
    ELITE_MANGUDAI = 561, 42, 135, 16458, False
    ELITE_ORGAN_GUN = 1003, 190, 1002, 16101, False
    ELITE_PLUMED_ARCHER = 765, 108, 764, 16685, False
    ELITE_RATTAN_ARCHER = 1131, 232, 1130, 16101, False
    ELITE_SAMURAI = 560, 44, 151, 16460, False
    ELITE_SHOTEL_WARRIOR = 1018, 195, 1017, 16101, False
    ELITE_SKIRMISHER = 6, 21, 100, 16087, False
    ELITE_STEPPE_LANCER = 1372, 274, 1373, 16746, False
    ELITE_TARKAN = 757, 105, 756, 16677, False
    ELITE_TEUTONIC_KNIGHT = 554, 45, 181, 16462, False
    ELITE_THROWING_AXEMAN = 531, 46, 157, 16461, False
    ELITE_TURTLE_SHIP = 832, 116, -1, 16106, False
    ELITE_WAR_ELEPHANT = 558, 43, 136, 16459, False
    ELITE_WAR_WAGON = 829, 117, 828, 16729, False
    ELITE_WOAD_RAIDER = 534, 47, 233, 16463, False
    FAST_FIRE_SHIP = 532, 85, -1, 16429, False
    FIRE_GALLEY = 1103, 203, -1, 16426, False
    FIRE_SHIP = 529, 86, -1, 16426, False
    FISHING_SHIP = 13, 24, -1, 16090, False
    FLAMETHROWER = 188, 144, 189, 16109, False
    FLAMING_CAMEL = 1263, 270, -1, 16734, False
    FURIOUS_THE_MONKEY_BOY = 860, 132, 861, 16763, False
    GALLEON = 442, 60, -1, 16309, False
    GALLEY = 539, 87, -1, 16436, False
    GBETO = 1013, 197, 1014, 16101, False
    GENITOUR = 1010, 201, 1011, 16417, False
    GENOESE_CROSSBOWMAN = 866, 133, 867, 16101, False
    GOAT = 1060, 200, 1061, 16061, False
    GOOSE = 1243, 265, 1244, 16061, False
    HALBERDIER = 359, 104, 502, 16409, False
    HAND_CANNONEER = 5, 22, 98, 16086, False
    HEAVY_CAMEL_RIDER = 330, 79, 495, 16417, False
    HEAVY_CAVALRY_ARCHER = 474, 71, 631, 16412, False
    HEAVY_CROSSBOWMAN = 493, 133, 867, 16101, False
    HEAVY_DEMOLITION_SHIP = 528, 83, -1, 16425, False
    HEAVY_PIKEMAN = 892, 136, 893, 16410, False
    HEAVY_SCORPION = 542, 89, 543, 16439, False
    HEAVY_SWORDSMAN = 76, 184, 99, 16077, False
    HORSE_A = 814, 112, 815, 16714, False
    HORSE_B = 1356, 112, 1357, 16714, False
    HORSE_C = 1602, 112, 1603, 16714, False
    HORSE_D = 1604, 112, 1605, 16714, False
    HORSE_E = 1606, 112, 1607, 16714, False
    HUSKARL = 41, 50, 62, 16104, False
    HUSSAR = 441, 103, 480, 16661, False
    IMPERIAL_CAMEL_RIDER = 207, 185, 300, 16417, False
    IMPERIAL_SKIRMISHER = 1155, 229, 1156, 16087, False
    INVISIBLE_OBJECT = 1291, -1, -1, 16000, False
    IROQUOIS_WARRIOR = 1374, 297, 1375, 16113, False
    JAGUAR_WARRIOR = 725, 110, 750, 16667, False
    JANISSARY = 46, 39, 107, 16105, False
    JUNK = 15, 211, -1, 16092, False
    KAMAYUK = 879, 97, 880, 16110, False
    KARAMBIT_WARRIOR = 1123, 233, 1124, 16101, False
    KESHIK = 1228, 251, 1229, 16101, False
    KHAN = 1275, 258, 1277, 16458, False
    KING = 434, 48, 435, 16301, False
    KIPCHAK = 1231, 252, 1232, 16108, False
    KNIGHT = 38, 1, 111, 16068, False
    KONNIK = 1225, 249, 1252, 16101, False
    KONNIK_DISMOUNTED = 1252, 250, 1257, 16411, False
    LEGIONARY = 1, 139, 2, 16669, False
    LEITIS = 1234, 253, 1235, 16101, False
    LIGHT_CAVALRY = 546, 91, 547, 16444, False
    LLAMA = 305, 156, 780, 16498, False
    LONG_SWORDSMAN = 77, 13, 180, 16081, False
    LONGBOAT = 250, 40, -1, 16106, False
    LONGBOWMAN = 8, 41, 115, 16107, False
    MAGYAR_HUSZAR = 869, 99, 870, 16101, False
    MAMELUKE = 282, 37, 44, 16103, False
    MAN_AT_ARMS = 75, 10, 154, 16080, False
    MANGONEL = 280, 27, 121, 16095, False
    MANGUDAI = 11, 42, 135, 16108, False
    MERCHANT = 1572, 346, 1573, 16301, False
    MILITIA = 74, 8, 152, 16079, False
    MISSIONARY = 775, 107, 776, 16691, False
    MONK = 125, 33, 134, 16099, False
    MONK_WITH_RELIC = 286, 33, 134, 16380, False
    NINJA = 1145, 299, 1147, 16110, False
    NORSE_WARRIOR = 361, 140, 362, 16081, False
    ONAGER = 550, 101, 675, 16448, False
    ORGAN_GUN = 1001, 190, 1002, 16101, False
    OX_CART = 1271, 263, 1272, 16100, False
    OX_WAGON = 1273, 264, 1274, 16100, False
    PALADIN = 569, 2, 570, 16471, False
    PENGUIN = 639, 157, 641, 16469, False
    PETARD = 440, 113, -1, 16660, False
    PHOTONMAN = 1577, 300, 1578, 16043, False
    PIG = 1245, 269, 1246, 16061, False
    PIKEMAN = 358, 11, 501, 16408, False
    PLUMED_ARCHER = 763, 108, 764, 16683, False
    PRIEST = 1023, 294, 1024, 16120, False
    PRIEST_WITH_RELIC = 1400, 33, 1024, 16380, False
    QUEEN = 1292, 168, 1328, 16302, False
    RATTAN_ARCHER = 1129, 232, 1130, 16101, False
    RELIC_CART = 1304, 295, 285, 16082, False
    ROYAL_JANISSARY = 52, 296, 1576, 16455, False
    SAMURAI = 291, 44, 151, 16110, False
    SCORPION = 279, 80, 149, 16096, False
    SCOUT_CAVALRY = 448, 64, 449, 16326, False
    SHARKATZOR = 1222, 352, -1, 16458, False
    SHEEP = 594, 96, 595, 16498, False
    SHOTEL_WARRIOR = 1016, 195, 1017, 16101, False
    SIEGE_ONAGER = 588, 102, 589, 16493, False
    SIEGE_RAM = 548, 73, 549, 16446, False
    SIEGE_TOWER = 1105, 212, 1107, 16445, False
    SKIRMISHER = 7, 20, 238, 16088, False
    SLINGER = 185, 143, 186, 16743, False
    SPEARMAN = 93, 31, 140, 16078, False
    STEPPE_LANCER = 1370, 273, 1371, 16746, False
    TARKAN = 755, 105, 756, 16675, False
    TEUTONIC_KNIGHT = 25, 45, 181, 16112, False
    THROWING_AXEMAN = 281, 46, 157, 16111, False
    TORCH_A_CONVERTABLE = 854, -1, -1, 16753, False
    TORCH_B_CONVERTABLE = 1377, -1, -1, 16753, False
    TRADE_CART_EMPTY = 128, 34, 178, 16100, False
    TRADE_CART_FULL = 204, 34, 205, 16100, False
    TRADE_COG = 17, 23, -1, 16089, False
    TRANSPORT_SHIP = 545, 95, -1, 16443, False
    TREBUCHET = 42, 28, 194, 16097, False
    TREBUCHET_PACKED = 331, 29, 735, 16381, False
    TURKEY = 833, 115, 834, 16733, False
    TURTLE_SHIP = 831, 116, -1, 16106, False
    TWO_HANDED_SWORDSMAN = 473, 12, 500, 16411, False
    VILLAGER_MALE = 83, 15, 224, 16121, False
    VILLAGER_FEMALE = 293, 16, 211, 16121, False
    VILLAGER_MALE_BUILDER = 118, 330, 225, 16122, False
    VILLAGER_FEMALE_BUILDER = 212, 329, 213, 16122, False
    VILLAGER_MALE_FARMER = 259, 332, 226, 16123, False
    VILLAGER_FEMALE_FARMER = 214, 331, 215, 16123, False
    VILLAGER_MALE_FISHERMAN = 56, 332, 58, 16499, False
    VILLAGER_FEMALE_FISHERMAN = 57, 331, 60, 16499, False
    VILLAGER_MALE_FORAGER = 120, 332, 353, 16402, False
    VILLAGER_FEMALE_FORAGER = 354, 331, 355, 16402, False
    VILLAGER_MALE_GOLD_MINER = 579, 334, 229, 16482, False
    VILLAGER_FEMALE_GOLD_MINER = 581, 333, 221, 16482, False
    VILLAGER_MALE_HUNTER = 122, 332, 227, 16124, False
    VILLAGER_FEMALE_HUNTER = 216, 331, 217, 16124, False
    VILLAGER_MALE_LUMBERJACK = 123, 339, 228, 16125, False
    VILLAGER_FEMALE_LUMBERJACK = 218, 338, 219, 16125, False
    VILLAGER_MALE_REPAIRER = 156, 330, 225, 16127, False
    VILLAGER_FEMALE_REPAIRER = 222, 329, 213, 16127, False
    VILLAGER_MALE_SHEPHERD = 592, 332, 593, 16496, False
    VILLAGER_FEMALE_SHEPHERD = 590, 331, 591, 16496, False
    VILLAGER_MALE_STONE_MINER = 124, 336, 229, 16126, False
    VILLAGER_FEMALE_STONE_MINER = 220, 335, 221, 16126, False
    VMDL = 206, 337, -1, 16656, False
    WAR_ELEPHANT = 239, 43, 136, 16109, False
    WAR_GALLEY = 21, 25, -1, 16091, False
    WAR_WAGON = 827, 117, 828, 16727, False
    WATER_BUFFALO = 1142, 224, 1143, 16175, False
    WOAD_RAIDER = 232, 47, 233, 16113, False
    XOLOTL_WARRIOR = 1570, 351, 1571, 16326, False
    BOARDER_GALLEY = 536, 82, -1, 16703, False
    HUSKARL_BARRACKS = 759, 50, 62, 16748, False
    ELITE_HUSKARL_BARRACKS = 761, 50, 62, 16748, False
    TARKAN_STABLE = 886, 105, 756, 16741, False
    ELITE_TARKAN_STABLE = 887, 105, 756, 16741, False
    GENITOUR_ORIGINAL = 583, 19, 152, 16663, False
    ELITE_GENITOUR_ORIGINAL = 596, 19, 152, 16665, False
    COUSTILLIER = 1655, 355, 1656, 16101, False
    ELITE_COUSTILLIER = 1657, 355, 1656, 16101, False
    FLEMISH_MILITIA = 1699, 354, 1664, 16735, False
    FLEMISH_MILITIA_MALE = 1663, 354, 1664, 16469, False
    FLEMISH_MILITIA_FEMALE = 1697, 354, 1698, 16469, False
    SERJEANT = 1658, 356, 1662, 16104, False
    ELITE_SERJEANT = 1659, 356, 1662, 16454, False
    SERJEANT_DONJON = 1660, 356, 1662, 16101, False
    ELITE_SERJEANT_DONJON = 1661, 356, 1662, 16101, False
    CONDOTTIERO_PLACEHOLDER = 184, 134, 883, 16679, False
    KONNIK_KREPOST = 1254, 249, 1252, 16101, False
    ELITE_KONNIK_KREPOST = 1255, 249, 1253, 16101, False
    ELITE_KIPCHAK_CUMAN_MERCENARIES = 1260, 252, 1232, 16730, False
    FOOD_WOOD_GOLD_TRICKLE = 1654, -1, -1, 16737, False
    CUMAN_DISABLED = 1261, 114, 1232, 16108, False
    OUTLW = 158, 17, 3, 16074, True
    RFARC = 571, 17, 572, 16474, False
    MONUMENT_RESOURCE_ENABLER = 1639, 17, 572, 16464, False
    HUSSITE_WAGON = 1704, 370, 1705, 16101, False
    ELITE_HUSSITE_WAGON = 1706, 370, 1705, 16101, False
    OBUCH = 1701, 369, 1702, 16104, False
    ELITE_OBUCH = 1703, 369, 1702, 16104, False
    WINGED_HUSSAR = 1707, 371, 1708, 16661, False
    HOUFNICE = 1709, 372, 1710, 16093, False
    URUMI_SWORDSMAN = 1735, 386, 1736, 16104, False
    MONK_WITH_TURKISH_RELIC = 309, 33, 134, 16361, False
    CRUSADER_KNIGHT = 1723, 377, 1724, 16730, False
    ELITE_URUMI_SWORDSMAN = 1737, 386, 1736, 16454, False
    RATHA_MELEE = 1738, 388, 1739, 16101, False
    ELITE_RATHA_MELEE = 1740, 388, 1739, 16101, False
    CHAKRAM_THROWER = 1741, 390, 1742, 16111, False
    ELITE_CHAKRAM_THROWER = 1743, 390, 1742, 16461, False
    GHULAM = 1747, 385, 1748, 16101, False
    ELITE_GHULAM = 1749, 385, 1748, 16101, False
    THIRISADAI = 1750, 387, -1, 16106, False
    SHRIVAMSHA_RIDER = 1751, 391, 1752, 16737, False
    ELITE_SHRIVAMSHA_RIDER = 1753, 391, 1752, 16737, False
    CAMEL_SCOUT = 1755, 392, 1668, 16416, False
    RATHA_RANGED = 1759, 389, 1760, 16727, False
    ELITE_RATHA_RANGED = 1761, 389, 1760, 16729, False
    ARMORED_ELEPHANT = 1744, 394, 1745, 16494, False
    SIEGE_ELEPHANT = 1746, 395, 1757, 16494, False



from AoE2ScenarioParser.helper.exceptions import EndOfFileError


class IncrementalGenerator:
    def __init__(self, name, file_content, progress=0):
        self.name = name
        self.file_content: bytes = file_content
        self.progress = progress

    @classmethod
    def from_file(cls, filepath: str):
        with open(filepath, 'rb') as f:
            file_content = f.read()
        return cls(filepath, file_content)

    def get_bytes(self, n: int, update_progress=True):
        if n <= 0:
            return b''
        result = self.file_content[self.progress:self.progress + n]
        if not result:
            raise EndOfFileError("End of file reached")
        if update_progress:
            self.progress += n
        return result

    def get_remaining_bytes(self):
        result = self.file_content[self.progress:]
        self.progress = len(self.file_content) - 1
        return result

    def __repr__(self):
        return f"[IncrementalGenerator] Name: {self.name}\n\tProgress: {self.progress}/{len(self.file_content)}"



from __future__ import annotations

from AoE2ScenarioParser.objects.aoe2_object import AoE2Object
from AoE2ScenarioParser.sections.retrievers.retriever_object_link import RetrieverObjectLink


class Variable(AoE2Object):
    """Object for handling a variable."""

    _link_list = [
        RetrieverObjectLink("variable_id", "Triggers", "variable_data[__index__].variable_id"),
        RetrieverObjectLink("name", "Triggers", "variable_data[__index__].variable_name"),
    ]

    def __init__(self, variable_id: int, name: str, **kwargs):
        self.variable_id = variable_id
        self.name = name

        super().__init__(**kwargs)



from typing import List

from AoE2ScenarioParser.objects.aoe2_object import AoE2Object
from AoE2ScenarioParser.objects.data_objects.unit import Unit
from AoE2ScenarioParser.sections.retrievers.retriever_object_link import RetrieverObjectLink


class PlayerUnits(AoE2Object):
    """Object for handling a tile in the map."""

    _link_list = [
        RetrieverObjectLink("unit_count", "Units", "players_units[__index__].unit_count"),
        RetrieverObjectLink("units", "Units", "players_units[__index__].units", process_as_object=Unit),
    ]

    def __init__(self, unit_count: int, units: List[Unit], **kwargs):
        super().__init__(**kwargs)

        self.unit_count: int = unit_count
        self.units: List[Unit] = units



from typing import List

from AoE2ScenarioParser.objects.data_objects.units.player_units import PlayerUnits
from AoE2ScenarioParser.objects.managers.unit_manager import UnitManager


class UnitManagerDE(UnitManager):

    def __init__(self, _player_units: List[PlayerUnits], **kwargs):
        super().__init__(_player_units, **kwargs)



from typing import List


def hash_list(lst: list):
    return hash(tuple(lst))


def list_changed(lst, lst_hash):
    return lst_hash != hash(tuple(lst))


def listify(var) -> list:
    """Always return item as list"""
    return var if type(var) is list else [var]


def update_order_array(order_array: List[int], supposed_length: int) -> None:
    """
    Update an order array.

    Args:
        order_array (List[int]): The order array like trigger.condition_order
        supposed_length (int): The length the array should be

    """
    actual_length = len(order_array)

    if actual_length > supposed_length:
        for i in range(supposed_length, actual_length):
            order_array.remove(i)
    elif supposed_length > actual_length:
        for i in range(supposed_length):
            if i not in order_array:
                order_array.append(i)


# Written by: Ned Batchelder @ https://stackoverflow.com/a/312464/7230293
def list_chuncks(lst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i:i + n]



from typing import Optional, List, Union, overload

from AoE2ScenarioParser.datasets.dataset_enum import dataset_or_value
from AoE2ScenarioParser.datasets.object_support import StartingAge, Civilization
from AoE2ScenarioParser.datasets.players import PlayerId
from AoE2ScenarioParser.datasets.trigger_lists import DiplomacyState
from AoE2ScenarioParser.helper.list_functions import listify
from AoE2ScenarioParser.objects.aoe2_object import AoE2Object


class Player(AoE2Object):
    """Object for handling all player information."""

    _object_attributes = [
        'player_id',
        'starting_age',
        'lock_civ',
        'food',
        'wood',
        'gold',
        'stone',
        'color',
        'active',
        'human',
        'civilization',
        'architecture_set',
    ]
    _object_attributes_non_gaia = [
        'population_cap',
        'diplomacy',
        'initial_camera_x',
        'initial_camera_y',
        'allied_victory',
        'disabled_techs',
        'disabled_buildings',
        'disabled_units',
        'base_priority',
        'tribe_name',
        'string_table_name_id',
    ]

    def __init__(
            self,
            player_id: int,
            starting_age: int,
            lock_civ: int,
            food: int,
            wood: int,
            gold: int,
            stone: int,
            color: int,
            active: bool,
            human: bool,
            civilization: int,
            architecture_set: int,

            # Optionals due to GAIA not having such value
            population_cap: Optional[int] = None,
            diplomacy: Optional[List[int]] = None,
            initial_camera_x: Optional[int] = None,
            initial_camera_y: Optional[int] = None,
            allied_victory: Optional[int] = None,
            disabled_techs: Optional[List[int]] = None,
            disabled_buildings: Optional[List[int]] = None,
            disabled_units: Optional[List[int]] = None,
            tribe_name: Optional[str] = None,
            base_priority: Optional[int] = None,
            string_table_name_id: Optional[int] = None,
            **kwargs
    ):
        super().__init__(**kwargs)

        self._player_id: int = player_id
        self._active: bool = active
        self.starting_age: int = dataset_or_value(StartingAge, starting_age)
        self.lock_civ: bool = bool(lock_civ)
        self.food: int = food
        self.wood: int = wood
        self.gold: int = gold
        self.stone: int = stone
        self.color: int = color
        self.human: bool = human
        self.civilization: Union[int, Civilization] = dataset_or_value(Civilization, civilization)
        self.architecture_set: Union[int, Civilization] = dataset_or_value(Civilization, architecture_set)

        # Optionals due to GAIA not having such value
        self.population_cap: Optional[int] = population_cap
        self.diplomacy: Optional[List[int]] = diplomacy
        self.initial_camera_x: Optional[int] = initial_camera_x
        self.initial_camera_y: Optional[int] = initial_camera_y
        self.allied_victory: Optional[bool] = bool(allied_victory) if allied_victory is not None else None
        self.disabled_techs: Optional[List[int]] = disabled_techs
        self.disabled_buildings: Optional[List[int]] = disabled_buildings
        self.disabled_units: Optional[List[int]] = disabled_units
        self.tribe_name: Optional[str] = tribe_name
        self.base_priority: Optional[int] = base_priority
        self.string_table_name_id: Optional[int] = string_table_name_id

    @property
    def player_id(self):
        """Read-only value of the player ID"""
        return self._player_id

    @property
    def active(self):
        return self._active

    @active.setter
    def active(self, value):
        raise ValueError("Cannot set active status of player directly, please use player_manager.active_players")

    @overload
    def set_player_diplomacy(self, players: List[Union[PlayerId, int]], diplomacy: DiplomacyState):
        ...

    @overload
    def set_player_diplomacy(self, player: Union[PlayerId, int], diplomacy: DiplomacyState):
        ...

    def set_player_diplomacy(self, *args):
        players: List[Union[PlayerId, int]] = listify(args[0])
        diplomacy: DiplomacyState = args[1]

        if self.player_id in players:
            raise ValueError("Cannot set diplomacy from and to the same player")

        for player in players:
            self.diplomacy[player - 1] = diplomacy

    def _get_object_attrs(self):
        attrs = self._object_attributes
        if self.player_id != PlayerId.GAIA:
            attrs.extend(self._object_attributes_non_gaia)
        return super()._get_object_attrs() + attrs



import sys

from AoE2ScenarioParser import settings
from AoE2ScenarioParser.helper.exceptions import WarningToError
from AoE2ScenarioParser.helper.string_manipulations import add_tabs

_color = {
    'end': '\033[0m',
    'black': '\033[30m',
    "red": "\033[31m",
    "green": "\033[32m",
    "yellow": "\033[33m",
    "blue": "\033[34m",
    "magenta": "\033[35m",
    "cyan": "\033[36m",
    "white": "\033[37m",
    "bright_black": "\033[90m",
    "bright_red": "\033[91m",
    "bright_green": "\033[92m",
    "bright_yellow": "\033[93m",
    "bright_blue": "\033[94m",
    "bright_magenta": "\033[95m",
    "bright_cyan": "\033[96m",
    "bright_white": "\033[97m",
}


def rprint(string="", replace=True, final=False) -> None:
    """
    Replaceable print, print lines which can be overwritten by the next

    Args:
        string (str): The string to print -> print(str)
        replace (bool): If this line should be replaced by the next if the next also has replace=True
        final (bool): If true, the next print is not able to replace this line. Used when this line replaces
            another line but should not be replaced by the next replace=True.

    Returns:
        None
    """
    if replace:
        sys.stdout.write('\r' + string)
        if final:
            print()
    else:
        print(string)


def s_print(string="", replace=True, final=False, color=None) -> None:
    """
    Status print, read rprint docstring for more info.
    Simple rprint wrapper with a check for the PRINT_STATUS_UPDATES setting.
    """
    if settings.PRINT_STATUS_UPDATES:
        if color is not None:
            string = color_string(string, color)
        rprint(string, replace, final)


def color_string(string: str, color: str) -> str:
    return _color[color] + string + _color['end']


def warn(string="") -> None:
    if settings.RAISE_ERROR_ON_WARNING:
        raise WarningToError(
            f"\n\nWarning Occurred with 'settings.THROW_ERROR_ON_WARNING' is True. \nWarning:\n{add_tabs(string, 1)}\n")
    if not settings.DISABLE_WARNINGS:
        print('\n' + color_string(string, 'bright_yellow'))



import json

from AoE2ScenarioParser.datasets.buildings import BuildingInfo
from AoE2ScenarioParser.datasets.heroes import HeroInfo
from AoE2ScenarioParser.datasets.other import OtherInfo
from AoE2ScenarioParser.datasets.units import UnitInfo

#
# Written by Alian713
#


with open("./full.json") as file:
    dataset = json.load(file)

_icon_ids = {}

_unit_icon = {}
for obj in UnitInfo:
    for unit in dataset["Civs"][0]["Units"]:
        if unit["ID"] == obj.ID:
            _unit_icon[str(obj.ID).split(".")[1]] = unit["IconID"]

_icon_ids["_unit_icon"] = _unit_icon

_building_icon = {}
for obj in BuildingInfo:
    for unit in dataset["Civs"][0]["Units"]:
        if unit["ID"] == obj.ID:
            _building_icon[str(obj.ID).split(".")[1]] = unit["IconID"]

_icon_ids["_building_icon"] = _building_icon

# _gaia_unit_icon = {}
# for obj in GaiaUnitId:
#     for unit in dataset["Civs"][0]["Units"]:
#         if unit["ID"] == obj:
#             _gaia_unit_icon[str(obj).split(".")[1]] = unit["IconID"]
#
# _icon_ids["_gaia_unit_icon"] = _gaia_unit_icon

# _gaia_building_icon = {}
# for obj in GaiaBuildingId:
#     for unit in dataset["Civs"][0]["Units"]:
#         if unit["ID"] == obj:
#             _gaia_building_icon[str(obj).split(".")[1]] = unit["IconID"]
#
# _icon_ids["_gaia_building_icon"] = _gaia_building_icon

_unit_other = {}
for obj in OtherInfo:
    for unit in dataset["Civs"][0]["Units"]:
        if unit["ID"] == obj.ID:
            _unit_other[str(obj.ID).split(".")[1]] = unit["IconID"]

_icon_ids["_unit_other"] = _unit_other

# _gaia_unit_other_icon = {}
# for obj in GaiaUnitOtherId:
#     for unit in dataset["Civs"][0]["Units"]:
#         if unit["ID"] == obj:
#             _gaia_unit_other_icon[str(obj).split(".")[1]] = unit["IconID"]
#
# _icon_ids["_gaia_unit_other_icon"] = _gaia_unit_other_icon

_hero_icon = {}
for obj in HeroInfo:
    for unit in dataset["Civs"][0]["Units"]:
        if unit["ID"] == obj.ID:
            _hero_icon[str(obj.ID).split(".")[1]] = unit["IconID"]

_icon_ids["_hero_icon"] = _hero_icon

with open("./icons.json", "w") as file:
    json.dump(_icon_ids, file, indent=4)

# Change json to print Python code


with open('./icons.json', 'r') as f:
    js = json.loads(f.read())

    for header in js:
        print(f"class {header}(IntEnum):")
        for key, v in js.get(header).items():
            print(f"    {key} = {v}")

        input("Press Enter to continue...")



from typing import Dict, TYPE_CHECKING, List, Union, Callable
from uuid import UUID

from AoE2ScenarioParser.datasets.conditions import attribute_presentation as condition_attribute_presentation
from AoE2ScenarioParser.datasets.effects import attribute_presentation as effect_attribute_presentation
from AoE2ScenarioParser.datasets.players import PlayerId, PlayerColorId
from AoE2ScenarioParser.datasets.support.info_dataset_base import InfoDatasetBase
from AoE2ScenarioParser.datasets.techs import TechInfo
from AoE2ScenarioParser.datasets.trigger_lists import DiplomacyState, Operation, AttackStance, UnitAIAction, \
    ButtonLocation, PanelLocation, TimeUnit, VisibilityState, DifficultyLevel, TechnologyState, Comparison, \
    ObjectAttribute, ObjectType, ObjectClass, TerrainRestrictions, HeroStatusFlag, BlastLevel, \
    DamageClass, Hotkey, ColorMood, ObjectState, ActionType, VictoryTimerType, Attribute, ProjectileSmartMode
from AoE2ScenarioParser.helper.helper import get_enum_from_unit_const
from AoE2ScenarioParser.helper.list_functions import listify
from AoE2ScenarioParser.helper.pretty_format import pretty_format_name
from AoE2ScenarioParser.helper.string_manipulations import q_str, trunc_string
from AoE2ScenarioParser.scenarios.scenario_store import getters

if TYPE_CHECKING:
    from AoE2ScenarioParser.objects.data_objects.unit import Unit

_store_error_displays: Dict[str, Dict[str, Callable[..., str]]] = {
    'triggers': {
        'invalid_reference': lambda id_=None: f"<<INVALID TRIGGER>>"
    },
    'variables': {
        'invalid_reference': lambda id_=None: f"<<INVALID VARIABLE>>"
    },
    'units': {
        'invalid_reference': lambda id_: f"<<INVALID UNIT>> ({id_})",
        'unknown': lambda const_: f"Unknown"
    }
}


def _format_trigger_id_representation(id_: int, uuid: UUID) -> str:
    if (name := getters.get_trigger(uuid, id_).name) is not None:
        return f"\"{trunc_string(name)}\""
    return _store_error_displays['triggers']['invalid_reference']()


def _format_variable_id_representation(id_: int, uuid: UUID) -> str:
    if (name := getters.get_variable_name(uuid, id_)) is not None:
        return f"\"{trunc_string(name)}\""
    return _store_error_displays['variables']['invalid_reference']()


def _format_unit_reference_representation(ref_id: Union[int, List[int]], uuid: UUID) -> str:
    def format_unit(u: 'Unit') -> str:
        enum_entry = u.unit_const
        if not issubclass(u.unit_const.__class__, InfoDatasetBase):
            enum_entry = get_enum_from_unit_const(u.unit_const)

        if enum_entry:
            name = pretty_format_name(enum_entry.name)
        else:
            name = _store_error_displays['units']['unknown'](u.unit_const)

        return f"{name} ({u.unit_const}) [P{u.player}, X{u.x}, Y{u.y}]"

    ids = listify(ref_id)
    retrieved_units = getters.get_units(uuid, ids)
    if retrieved_units is not None:
        units, invalids = retrieved_units
        unit_count = len(units) + len(invalids)
        formatted = '\n\t'.join([
            f"{i}: {format_unit(unit)} ({unit.reference_id})" for i, unit in enumerate(units)
        ])
        formatted += '\n\t'.join([
            _store_error_displays['units']['invalid_reference'](invalid_id) for invalid_id in invalids
        ])
        s = "s" if unit_count != 1 else ""
        return f"{unit_count} unit{s}:\n\t{formatted}"
    return f"<<INVALID UNITS>> ({ref_id})"


_datasets = {
    "DiplomacyState": DiplomacyState,
    "Operation": Operation,
    "AttackStance": AttackStance,
    "UnitAIAction": UnitAIAction,
    "ButtonLocation": ButtonLocation,
    "PanelLocation": PanelLocation,
    "TimeUnit": TimeUnit,
    "VisibilityState": VisibilityState,
    "DifficultyLevel": DifficultyLevel,
    "TechnologyState": TechnologyState,
    "Comparison": Comparison,
    "ObjectAttribute": ObjectAttribute,
    "Attribute": Attribute,
    "ObjectType": ObjectType,
    "ObjectClass": ObjectClass,
    "TerrainRestrictions": TerrainRestrictions,
    "HeroStatusFlag": HeroStatusFlag,
    "BlastLevel": BlastLevel,
    "SmartProjectile": ProjectileSmartMode,
    "DamageClass": DamageClass,
    "Hotkey": Hotkey,
    "ColorMood": ColorMood,
    "ObjectState": ObjectState,
    "ActionType": ActionType,
    "VictoryTimerType": VictoryTimerType,
}

_other = {
    'bool': lambda v: str(bool(v)),
    'PlayerId': lambda p: f"Player {PlayerId(p).name.capitalize()}",
    'PlayerColorId': lambda p: f"{PlayerColorId(p + 1).name.capitalize()}",
    'str': q_str,
}

_combined_info_datasets = [
    "UnitInfo",
    "BuildingInfo",
    "OtherInfo",
    "HeroInfo"
]
_other_info_datasets = {
    "TechInfo": TechInfo
}
_store_references = {
    "TriggerId": _format_trigger_id_representation,
    "Unit": _format_unit_reference_representation,
    "Unit[]": _format_unit_reference_representation,
    "VariableId": _format_variable_id_representation,
}


def transform_effect_attr_value(effect_type, attr, val, uuid):
    return transform_attr_value('e', effect_type, attr, val, uuid)


def transform_condition_attr_value(condition_type, attr, val, uuid):
    return transform_attr_value('c', condition_type, attr, val, uuid)


def transform_attr_value(ce, type_, attr, val, uuid):
    source: Dict[int, Dict[str, str]]
    source = condition_attribute_presentation if ce == 'c' else effect_attribute_presentation

    representation = get_presentation_value(attr, source, type_)

    if representation == "":
        return val

    value_representation = transform_value_by_representation(representation, val, uuid)
    return value_representation


def get_presentation_value(key, source, type_):
    if type_ not in source:
        return None
    if key in source[type_]:
        return source[type_][key]
    return source[-1][key]


def transform_value_by_representation(representation, value, uuid):
    unknown = "Unknown", False

    format_value_repr = True
    suffix_original_value = True

    try:
        if representation in _datasets:
            value_representation = _datasets[representation](value).attribute_presentation()

        elif representation in _combined_info_datasets:
            enum_entry = get_enum_from_unit_const(value)
            if enum_entry is not None:
                value_representation = enum_entry.name
            else:
                value_representation, format_value_repr = unknown

        elif representation in _other_info_datasets:
            value_representation = _other_info_datasets[representation].from_id(value).name

        elif representation in _store_references:
            value_representation = _store_references[representation](value, uuid)
            format_value_repr = False

            if representation in ["Unit", "Unit[]"]:
                suffix_original_value = False

        elif representation in _other:
            value_representation = _other[representation](value)
            format_value_repr = False

            if representation == "str":
                suffix_original_value = False
        else:
            raise ValueError(f"Unknown representation: '{representation}'")
    except (KeyError, ValueError):
        value_representation, format_value_repr = unknown

    if format_value_repr:
        value_representation = pretty_format_name(value_representation)
    if suffix_original_value:
        value_representation += f" ({value})"
    return value_representation



from __future__ import annotations

from typing import List, Union, Tuple

from AoE2ScenarioParser.datasets.players import PlayerId
from AoE2ScenarioParser.objects.aoe2_object import AoE2Object
from AoE2ScenarioParser.objects.data_objects.unit import Unit
from AoE2ScenarioParser.objects.data_objects.units.player_units import PlayerUnits
from AoE2ScenarioParser.objects.support.tile import Tile
from AoE2ScenarioParser.objects.support.uuid_list import UuidList
from AoE2ScenarioParser.sections.retrievers.retriever_object_link import RetrieverObjectLink


class UnitManager(AoE2Object):
    """Manager of the everything trigger related."""

    _link_list = [
        RetrieverObjectLink("_player_units", "Units", "players_units", process_as_object=PlayerUnits),
        RetrieverObjectLink("next_unit_id", "DataHeader", "next_unit_id_to_place")
    ]

    def __init__(
            self,
            _player_units: List[PlayerUnits],
            next_unit_id: int,
            **kwargs
    ):
        super().__init__(**kwargs)

        self.units = [pu.units for pu in _player_units]
        self.reference_id_generator = create_id_generator(next_unit_id)

    @property
    def next_unit_id(self):
        return self.get_new_reference_id()

    @property
    def units(self):
        return self._units

    @units.setter
    def units(self, value: List[List[Unit]]):
        def _raise():
            raise ValueError("Units should be list with a maximum of 9 sub lists, example: [[Unit], [Unit, Unit], ...]")

        if len(value) > 9:
            _raise()
        elif len(value) < 9:
            value.extend([[] for _ in range(9 - len(value))])

        self._units = UuidList(self._host_uuid, value)

    def update_unit_player_values(self):
        """Function to update all player values in all units. Useful when units are moved manually (in mass)."""
        for player in PlayerId.all():
            for unit in self.units[player]:
                unit._player = player

    def add_unit(self,
                 player: Union[int, PlayerId],
                 unit_const: int,
                 x: float = 0,
                 y: float = 0,
                 z: float = 0,
                 rotation: float = 0,
                 garrisoned_in_id: int = -1,
                 animation_frame: int = 0,
                 status: int = 2,
                 reference_id: int = None,
                 tile: Tile | Tuple[int, int] = None,
                 ) -> Unit:
        """
        Adds a unit to the scenario.

        Args:
            player: The player the unit belongs to.
            unit_const: Defines what unit you're placing. The IDs used in the unit/buildings dataset.
            x: The x location in the scenario.
            y: The y location in the scenario.
            z: The z (height) location in the scenario.
            rotation: The rotation of the unit.
            garrisoned_in_id: The reference_id of another unit this unit is garrisoned in.
            animation_frame: The animation frame of the unit.
            status: Unknown - Always 2. 0-6 no difference (?) | 7-255 makes it disappear. (Except from the mini-map)
            reference_id: The reference ID of this unit. Normally added automatically. Used for garrisoning or reference
                in triggers
            tile: An object that represents a tile on the map. Replaces parameters x and y. Also automatically adds
                .5 to both ints to place the unit centered on the tile.

        Returns:
            The Unit created
        """
        if reference_id is None:
            reference_id = self.get_new_reference_id()

        unit = Unit(
            player=player,
            x=x if tile is None else (tile[0] + .5),
            y=y if tile is None else (tile[1] + .5),
            z=z,
            reference_id=reference_id,
            unit_const=unit_const,
            status=status,
            rotation=rotation,
            initial_animation_frame=animation_frame,
            garrisoned_in_id=garrisoned_in_id,
            host_uuid=self._host_uuid
        )

        self.units[player].append(unit)
        return unit

    def get_player_units(self, player: Union[int, PlayerId]) -> List[Unit]:
        """
        Returns a list of UnitObjects for the given player.

        Raises:
            ValueError: If player is not between 0 (GAIA) and 8 (EIGHT)
        """
        if not 0 <= player <= 8:
            raise ValueError("Player must have a value between 0 and 8")
        return self.units[player]

    def get_all_units(self) -> List[Unit]:
        units = []
        for player_units in self.units:
            units += player_units
        return units

    def filter_units_by_const(self,
                              unit_consts: List[int],
                              blacklist: bool = False,
                              player_list: List[Union[int, PlayerId]] = None,
                              unit_list: List[Unit] = None) -> List[Unit]:
        """
        Filter unit on their unit_const value.

        Args:
            unit_consts (List[int]): The constants to filter with
            blacklist (bool): Use the given constant list as blacklist instead of whitelist
            player_list (List[int]): A list of players to filter from. If not used, all players are used.
            unit_list (List[Unit]): A set of units to filter from. If not used, all units are used.

        Returns:
            A list of units
        """
        if unit_list is None:
            unit_list = self.get_all_units()
        if player_list is not None:
            unit_list = [unit for unit in unit_list if unit.player in player_list]

        # Both return statements can be combined using: ((unit.unit_const in unit_consts) != blacklist)
        # But splitting them helps performance (not checking against blacklist for each entry)
        if not blacklist:
            return [unit for unit in unit_list if unit.unit_const in unit_consts]
        return [unit for unit in unit_list if unit.unit_const not in unit_consts]

    def get_units_in_area(self,
                          x1: float = None,
                          y1: float = None,
                          x2: float = None,
                          y2: float = None,
                          tile1: Tile = None,
                          tile2: Tile = None,
                          unit_list: List[Unit] = None,
                          players: List[Union[int, PlayerId]] = None,
                          ignore_players: List[PlayerId] = None):
        """
        Returns all units in the square with left corner (x1, y1) and right corner (x2, y2). Both corners inclusive.

        Args:
            x1: The X location of the left corner
            y1: The Y location of the left corner
            x2: The X location of the right corner
            y2: The Y location of the right corner
            tile1: The x,y location of the 1st corner as Tile Object
            tile2: The x,y location of the 2nd corner as Tile Object
            unit_list: (Optional) A list of units (Defaults to all units in the map, including GAIA (Trees etc.)
            players: (Optional) A list of Players which units need to be selected from the selected area
            ignore_players: (Optional) A list of Players which units need to be ignored from the selected area

        Raises:
            ValueError: if not all 4 (x1, y1, x2 and y2) are used simultaneously.
                Or if both (tile1 and tile2) are not used simultaneously.
                Or if any of the 4 (x1, y1, x2, y2) is used together with any of (tile1, tile2). Use one or the other.
                Or if players and ignore_players are used simultaneously.
        """
        if (x1 is not None or y1 is not None or x2 is not None or y2 is not None) and any([tile1, tile2]):
            raise ValueError("Cannot use both x1,y1,x2,y2 notation and tile1,tile2 notation at the same time")
        if (x1 is not None or y1 is not None or x2 is not None or y2 is not None) and \
                (x1 is None or y1 is None or x2 is None or y2 is None):
            raise ValueError("Cannot use some but not all from x1,y1,x2,y2.")
        if (not all([tile1, tile2])) and any([tile1, tile2]):
            raise ValueError("Cannot use one from tile1, tile2. Use both.")
        if players is not None and ignore_players is not None:
            raise ValueError("Cannot use both whitelist (players) and blacklist (ignore_players) at the same time")

        if tile1:
            x1 = tile1.x
            y1 = tile1.y
            x2 = tile2.x
            y2 = tile2.y
        else:
            # Inclusive selection
            x2 += 1
            y2 += 1

        if players is not None:
            players = players
        elif ignore_players is not None:
            players = [p for p in PlayerId if p not in ignore_players]
        else:
            players = [p for p in PlayerId]

        if unit_list is None:
            unit_list = self.get_all_units()

        return [unit for unit in unit_list if x1 <= unit.x <= x2 and y1 <= unit.y <= y2 and unit.player in players]

    @staticmethod
    def change_ownership(unit: Unit | List[Unit], to_player: Union[int, PlayerId]) -> None:
        """
        Changes a unit's ownership to the given player.

        Args:
            unit: The unit object which ownership will be changed
            to_player: The player that'll get ownership over the unit (using PlayerId enum)
        """
        if isinstance(unit, list):
            for u in unit:
                u.player = to_player
        else:
            unit.player = to_player

    def get_new_reference_id(self) -> int:
        """
        Get a new ID each time the function is called. Starting from the current highest ID.

        Returns:
            The newly generator ID
        """
        return next(self.reference_id_generator)

    def find_highest_reference_id(self) -> int:
        """
        Find the highest ID in the map. Searches through all units for the highest ID.

        Returns:
            The highest ID in the map
        """
        highest_id = 0  # If no units, default to 0
        for player in PlayerId.all():
            for unit in self.units[player]:
                highest_id = max(highest_id, unit.reference_id)
        return highest_id

    def remove_unit(self, reference_id: int = None, unit: Unit = None) -> None:
        """
        Removes a unit. Please note that `unit=...` is a lot faster than `reference_id=...` due to reference_id having
        to search through all units on the map. And unit has an ownership (player) attribute which is used for knowing
        which list to remove the unit from.

        Args:
            reference_id (int): The id of the unit. Note that this is NOT a unit constant (So NOT: UnitInfo.ARCHER)
            unit (Unit): The Unit object to be removed.
        """
        if reference_id is not None and unit is not None:
            raise ValueError("Cannot use both unit_ref_id and unit arguments. Use one or the other.")
        if reference_id is None and unit is None:
            raise ValueError("Both unit_ref_id and unit arguments were unused. Use one.")

        if reference_id is not None:
            for player in range(0, 9):
                for i, unit in enumerate(self.units[player]):
                    if unit.reference_id == reference_id:
                        del self.units[player][i]
                        return
        elif unit is not None:
            self.units[unit.player].remove(unit)

    def remove_eye_candy(self) -> None:
        eye_candy_ids = [1351, 1352, 1353, 1354, 1355, 1358, 1359, 1360, 1361, 1362, 1363, 1364, 1365, 1366]
        self.units[0] = [gaia_unit for gaia_unit in self.units[0] if gaia_unit.unit_const not in eye_candy_ids]

    # ###############################################################################################
    # ################################# Functions for reconstruction ################################
    # ###############################################################################################

    @property
    def _player_units(self):
        player_units = []
        for i in range(9):
            units = self.get_player_units(i)
            player_units.append(PlayerUnits(unit_count=len(units), units=units))

        return UuidList(self._host_uuid, player_units)


def create_id_generator(start_id: int):
    """
    Create generator for increasing value

    Args:
        start_id (int): The id to start returning

    Returns:
        A generator which will return a +1 ID value for each time called with next.
    """
    while True:
        yield start_id
        start_id += 1



import json
import uuid
import zlib
from pathlib import Path
from typing import Union, Dict

import AoE2ScenarioParser.datasets.conditions as conditions
import AoE2ScenarioParser.datasets.effects as effects
from AoE2ScenarioParser import settings
from AoE2ScenarioParser.helper.exceptions import InvalidScenarioStructureError, UnknownScenarioStructureError, \
    UnknownStructureError
from AoE2ScenarioParser.helper.incremental_generator import IncrementalGenerator
from AoE2ScenarioParser.helper.printers import s_print
from AoE2ScenarioParser.helper.string_manipulations import create_textual_hex
from AoE2ScenarioParser.helper.version_check import python_version_check
from AoE2ScenarioParser.objects.aoe2_object_manager import AoE2ObjectManager
from AoE2ScenarioParser.objects.managers.map_manager import MapManager
from AoE2ScenarioParser.objects.managers.player_manager import PlayerManager
from AoE2ScenarioParser.objects.managers.trigger_manager import TriggerManager
from AoE2ScenarioParser.objects.managers.unit_manager import UnitManager
from AoE2ScenarioParser.scenarios.scenario_store import store
from AoE2ScenarioParser.scenarios.support.object_factory import ObjectFactory
from AoE2ScenarioParser.scenarios.support.scenario_actions import ScenarioActions
from AoE2ScenarioParser.sections.aoe2_file_section import AoE2FileSection


class AoE2Scenario:
    @property
    def trigger_manager(self) -> TriggerManager:
        return self._object_manager.managers['Trigger']

    @property
    def unit_manager(self) -> UnitManager:
        return self._object_manager.managers['Unit']

    @property
    def map_manager(self) -> MapManager:
        return self._object_manager.managers['Map']

    @property
    def player_manager(self) -> PlayerManager:
        return self._object_manager.managers['Player']

    def __init__(self, source_location):
        self.source_location = source_location

        self.read_mode = None
        self.scenario_version = "???"
        self.game_version = "???"
        self.structure = {}
        self.sections: Dict[str, AoE2FileSection] = {}
        self._object_manager: Union[AoE2ObjectManager, None] = None

        # Used in debug functions
        self._file = None
        self._file_header = None
        self._decompressed_file_data = None

        self.uuid = uuid.uuid4()
        store.register_scenario(self)

        self.new = ObjectFactory(self.uuid)
        self.actions = ScenarioActions(self.uuid)

    @classmethod
    def from_file(cls, filename, game_version):
        python_version_check()

        s_print(f"\nReading file: '{filename}'", final=True, color="magenta")
        s_print("Reading scenario file...")
        igenerator = IncrementalGenerator.from_file(filename)
        s_print("Reading scenario file finished successfully.", final=True)

        scenario = cls(filename)
        scenario.read_mode = "from_file"
        scenario.game_version = game_version
        scenario.scenario_version = get_file_version(igenerator)

        # Log game and scenario version
        s_print("\n############### Attributes ###############", final=True, color="blue")
        s_print(f">>> Game version: '{scenario.game_version}'", final=True, color="blue")
        s_print(f">>> Scenario version: {scenario.scenario_version}", final=True, color="blue")
        s_print("##########################################", final=True, color="blue")

        s_print(f"\nLoading scenario structure...")
        initialise_version_dependencies(scenario.game_version, scenario.scenario_version)
        scenario._load_structure()
        s_print(f"Loading scenario structure finished successfully.", final=True)

        # scenario._initialize(igenerator)
        s_print("Parsing scenario file...", final=True)
        scenario._load_header_section(igenerator)
        scenario._load_content_sections(igenerator)
        s_print(f"Parsing scenario file finished successfully.", final=True)

        scenario._object_manager = AoE2ObjectManager(scenario.uuid)
        scenario._object_manager.setup()

        return scenario

    def _load_structure(self):
        if self.game_version == "???" or self.scenario_version == "???":
            raise ValueError("Both game and scenario version need to be set to load structure")
        self.structure = get_structure(self.game_version, self.scenario_version)

    def _load_header_section(self, raw_file_igenerator: IncrementalGenerator):
        header = self._create_and_load_section('FileHeader', raw_file_igenerator)
        self._add_to_sections(header)

    def _load_content_sections(self, raw_file_igenerator: IncrementalGenerator):
        self._decompressed_file_data = decompress_bytes(raw_file_igenerator.get_remaining_bytes())

        data_igenerator = IncrementalGenerator(name='Scenario Data', file_content=self._decompressed_file_data)

        for section_name in self.structure.keys():
            if section_name == "FileHeader":
                continue
            try:
                section = self._create_and_load_section(section_name, data_igenerator)
                self._add_to_sections(section)
            except (ValueError, TypeError) as e:
                print(f"\n[{e.__class__.__name__}] AoE2Scenario.parse_file: \n\tSection: {section_name}\n")
                self.write_error_file(trail_generator=data_igenerator)
                raise e

    def _create_and_load_section(self, name, igenerator):
        s_print(f"\t🔄 Parsing {name}...", color="yellow")
        section = AoE2FileSection.from_structure(name, self.structure.get(name), self.uuid)
        s_print(f"\t🔄 Gathering {name} data...", color="yellow")
        section.set_data_from_generator(igenerator)
        s_print(f"\t✔ {name}", final=True, color="green")
        return section

    def _add_to_sections(self, section):
        self.sections[section.name] = section

    def remove_store_reference(self):
        """
        Removes the reference to this scenario object from the scenario store. Useful (~a must) when reading many
        scenarios in a row without needing earlier ones. Python likes to take up a lot of memory.
        Removing all references to an object will cause the memory to be cleared up.

        **Please note:** When using this function it's important to remove all other references to the scenario.
        So if save it in a dict or list, remove it from it.
        If you have variables referencing this scenario that you won't need anymore (and won't overwrite) delete them
        using: `del varname`.
        """
        store.remove_scenario(self.uuid)

    """ ##########################################################################################
    ####################################### Write functions ######################################
    ########################################################################################## """

    def write_to_file(self, filename, skip_reconstruction=False):
        """
        Write the scenario to a new file

        Args:
            filename (str): The location to write the file to
            skip_reconstruction (bool): If reconstruction should be skipped. If true, this will ignore all changes made
                using the managers (For example all changes made using trigger_manager).
        """
        self._write_from_structure(filename, skip_reconstruction)

    def _write_from_structure(self, filename, skip_reconstruction=False):
        if not settings.DISABLE_ERROR_ON_OVERWRITING_SOURCE and self.source_location == filename:
            raise ValueError(
                "Overwriting the source scenario file is disallowed. This behaviour can be enabled in the settings file.")
        if not skip_reconstruction:
            self._object_manager.reconstruct()

        s_print("\nFile writing from structure started...", final=True)
        binary = _get_file_section_data(self.sections.get('FileHeader'))

        binary_list_to_be_compressed = []
        for file_part in self.sections.values():
            if file_part.name == "FileHeader":
                continue
            binary_list_to_be_compressed.append(_get_file_section_data(file_part))

        compressed = compress_bytes(b''.join(binary_list_to_be_compressed))

        with open(filename, 'wb') as f:
            f.write(binary + compressed)

        s_print("File writing finished successfully.", final=True)
        s_print(f"File successfully written to: '{filename}'", color="magenta", final=True)

    def write_error_file(self, filename="error_file.txt", trail_generator=None):
        self._debug_byte_structure_to_file(filename=filename, trail_generator=trail_generator)

    """ #############################################
    ################ Debug functions ################
    ############################################# """

    def _debug_write_from_source(self, filename, datatype, write_bytes=True):
        """This function is used as a test debugging writing. It writes parts of the read file to the filesystem."""
        print("File writing from source started with attributes " + datatype + "...")
        file = open(filename, "wb" if write_bytes else "w")
        selected_parts = []
        for t in datatype:
            if t == "f":
                selected_parts.append(self._file)
            elif t == "h":
                selected_parts.append(self._file_header)
            elif t == "d":
                selected_parts.append(self._decompressed_file_data)
        parts = None
        for part in selected_parts:
            if parts is None:
                parts = part
                continue
            parts += part
        file.write(parts if write_bytes else create_textual_hex(parts.hex()))
        file.close()
        print("File writing finished successfully.")

    def _debug_byte_structure_to_file(self, filename, trail_generator: IncrementalGenerator = None, commit=False):
        """ Used for debugging - Writes structure from read file to the filesystem in a easily readable manner. """
        if commit and hasattr(self, '_object_manager'):
            self._object_manager.reconstruct()

        s_print("\nWriting structure to file...", final=True)
        with open(filename, 'w', encoding=settings.MAIN_CHARSET) as f:
            result = []
            for section in self.sections.values():
                s_print(f"\t🔄 Writing {section.name}...")
                result.append(section.get_byte_structure_as_string())
                s_print(f"\t✔ {section.name}", final=True)

            if trail_generator is not None:
                s_print("\tWriting trail...")
                trail = trail_generator.get_remaining_bytes()

                result.append(f"\n\n{'#' * 27} TRAIL ({len(trail)})\n\n")
                result.append(create_textual_hex(trail.hex(), space_distance=2, enter_distance=24))
                s_print("\tWriting trail finished successfully.", final=True)

            f.write(''.join(result))
        s_print("Writing structure to file finished successfully.", final=True)


def initialise_version_dependencies(game_version, scenario_version):
    condition_json = get_version_dependant_structure_file(game_version, scenario_version, "conditions")

    for condition_id, structure in condition_json.items():
        condition_id = int(condition_id)

        if condition_id == -1:
            conditions.attribute_presentation[condition_id] = structure['attribute_presentation']
            continue

        conditions.condition_names[condition_id] = structure['name']
        conditions.default_attributes[condition_id] = structure['default_attributes']
        conditions.attributes[condition_id] = structure['attributes']
        conditions.attribute_presentation[condition_id] = structure.get('attribute_presentation', {})

    effect_json = get_version_dependant_structure_file(game_version, scenario_version, "effects")

    for effect_id, structure in effect_json.items():
        effect_id = int(effect_id)

        if effect_id == -1:
            effects.attribute_presentation[effect_id] = structure['attribute_presentation']
            continue

        effects.effect_names[effect_id] = structure['name']
        effects.default_attributes[effect_id] = structure['default_attributes']
        effects.attributes[effect_id] = structure['attributes']
        effects.attribute_presentation[effect_id] = structure.get('attribute_presentation', {})


def _get_file_section_data(file_section: AoE2FileSection):
    s_print(f"\t🔄 Reconstructing {file_section.name}...", color="yellow")
    value = file_section.get_data_as_bytes()
    s_print(f"\t✔ {file_section.name}", final=True, color="green")
    return value


def get_file_version(generator: IncrementalGenerator):
    """Get first 4 bytes of a file, which contains the version of the scenario"""
    return generator.get_bytes(4, update_progress=False).decode('ASCII')


def decompress_bytes(file_content):
    return zlib.decompress(file_content, -zlib.MAX_WBITS)


def compress_bytes(file_content):
    # https://stackoverflow.com/questions/3122145/zlib-error-error-3-while-decompressing-incorrect-header-check/22310760#22310760
    deflate_obj = zlib.compressobj(9, zlib.DEFLATED, -zlib.MAX_WBITS)
    compressed = deflate_obj.compress(file_content) + deflate_obj.flush()
    return compressed


def get_version_directory_path() -> Path:
    return Path(__file__).parent.parent / 'versions'


def get_version_dependant_structure_file(game_version: str, scenario_version: str, name: str) -> dict:
    try:
        vdir = get_version_directory_path()
        with (vdir / game_version / f'v{scenario_version}' / f'{name}.json').open() as structure_file:
            return json.load(structure_file)
    except FileNotFoundError:  # Unsupported version
        v = f"{game_version}:{scenario_version}"
        raise UnknownStructureError(f"The structure {name} could not be found with: {v}")


def get_structure(game_version, scenario_version) -> dict:
    try:
        vdir = get_version_directory_path()
        with (vdir / game_version / f'v{scenario_version}' / 'structure.json').open() as structure_file:
            structure = json.load(structure_file)

            if "FileHeader" not in structure.keys():
                raise InvalidScenarioStructureError(f"First section in structure should always be FileHeader.")
            return structure
    except FileNotFoundError:  # Unsupported version
        v = f"{game_version}:{scenario_version}"
        raise UnknownScenarioStructureError(f"The version {v} is not supported by AoE2ScenarioParser. :(") from None



NOTIFY_UNKNOWN_BYTES = True
"""Show a notification of extra bytes being available at the end of the file, so you can notify the maintainer."""

# Reading / Writing settings
PRINT_STATUS_UPDATES = True
"""If status updates of what is being read and written should be printed to console or not."""
DISABLE_ERROR_ON_OVERWRITING_SOURCE = False
"""Disable the error being raised when overwriting source scenario."""

# Warning related settings
DISABLE_WARNINGS = False
"""Disable all warnings thrown by the printers.warn function."""
DISABLE_VERSION_WARNINGS = False
"""Disable warnings about python versions"""
RAISE_ERROR_ON_WARNING = False
"""Raise an error when a warning should be shown. Handy for debugging, shows you the stacktrace."""

# Charset settings
MAIN_CHARSET = "utf-8"
"""The charset used to decode the text in the scenario. If it fails, will try the settings.FALLBACK_CHARSET"""
FALLBACK_CHARSET = "latin-1"
"""The charset used to decode the text in the scenario when the settings.MAIN_CHARSET fails."""

# Scenario construction
IGNORE_WRITING_ERRORS = False
"""Ignore possible errors raised while writing, useful for doing scenario debugging"""



from typing import Any


class TriggerComponent:
    hidden_attribute = ''
    """The attribute that should not be shown in this component. Used to hide the identifier in printing jobs"""

    def get_content_as_string(self) -> str:
        """Generate a string representation of this trigger component"""
        raise NotImplementedError("This method has not been implemented yet.")

    def should_be_displayed(self, attr: str, val: Any) -> bool:
        """
        Check if a given attribute with a given value should be printed or not

        Args:
            attr: The attribute as string
            val: The value found with the attribute

        Returns:
            A boolean if the attr, val combination should be printed
        """
        if val in [[], [-1], [''], "", " ", -1]:
            return False
        if attr == self.hidden_attribute:
            return False
        return True



from uuid import UUID

from AoE2ScenarioParser.objects.support.area import Area


class ObjectFactory:
    def __init__(self, uuid: UUID) -> None:
        super().__init__()

        self._uuid = uuid

    def area(self) -> Area:
        """Return an area map linked to the corresponding scenario"""
        return Area(uuid=self._uuid)



from typing import List

from AoE2ScenarioParser import settings
from AoE2ScenarioParser.helper.bytes_conversions import bytes_to_int
from AoE2ScenarioParser.helper.exceptions import EndOfFileError, UnsupportedAttributeError
from AoE2ScenarioParser.helper.incremental_generator import IncrementalGenerator

attributes = ['on_refresh', 'on_construct', 'on_commit']


def vorl(retriever, value):
    """
    Variable or List

    Args:
        retriever (Retriever): The retriever
        value (List[Any]): A value to be put in the retriever

    Returns:
        The given list or the value inside it
    """
    if retriever.datatype.repeat != 1:
        return value

    if retriever.is_list is not None:
        return value if retriever.is_list else value[0]

    # Fallback to length check
    if len(value) == 1:
        return value[0]

    return value


def retrieve_bytes(igenerator: IncrementalGenerator, retriever) -> List[bytes]:
    """
    Retrieves the bytes belonging to this retriever.

    Args:
        igenerator (IncrementalGenerator): The generator to return the bytes from
        retriever (Retriever): The retriever holding the bytes

    Returns:
        The corresponding bytes in a list. When 4 int8s need to retrieved, the list will have a length of 4 where each
            entry is 1 byte each.
    """
    var_type, var_len = retriever.datatype.type_and_length
    retrieved_bytes = []

    try:
        for i in range(retriever.datatype.repeat):
            if var_type != "str":  # (Signed) ints, floats, chars, plain bytes etc.
                retrieved_bytes.append(igenerator.get_bytes(var_len))
            else:  # String, Stored as: (signed int (n), string (string_length = n))
                int_bytes = igenerator.get_bytes(var_len)
                string_length = bytes_to_int(int_bytes, signed=True)
                string_bytes = igenerator.get_bytes(string_length)
                retrieved_bytes.append(int_bytes + string_bytes)
    except EndOfFileError:
        # END_OF_FILE_MARK retriever should always go in here
        if is_end_of_file_mark(retriever):
            retriever.datatype.repeat = 0
            return []

    # If more bytes present in the file after END_OF_FILE_MARK
    handle_end_of_file_mark(igenerator, retriever)

    return retrieved_bytes


def is_end_of_file_mark(retriever) -> bool:
    """Returns true if the retriever is the __END_OF_FILE_MARK__ retriever else false"""
    return retriever.name == "__END_OF_FILE_MARK__"


def handle_end_of_file_mark(igenerator, retriever) -> None:
    """
    Print message when the END_OF_FILE_MARK is reached and more bytes are present.\n
    You can disable this check (and thereby this message) using:\n
    ``>> from AoE2ScenarioParser import settings``\n
    ``>> settings.NOTIFY_UNKNOWN_BYTES = False``

    Args:
        igenerator (IncrementalGenerator): The generator to check if more bytes are present
        retriever (Retriever): The retriever to check if it's the end of file mark

    Returns:
        None
    """
    if is_end_of_file_mark(retriever) and settings.NOTIFY_UNKNOWN_BYTES:
        retrieved_bytes = igenerator.get_remaining_bytes()
        print("\n\n" + "\n".join([
            "The file being read has more bytes than anticipated.",
            "Please notify me (MrKirby/KSneijders) about this message!",
            "This will help with understanding more parts of scenario files! Thanks in advance!",
            "You can contact me using:",
            "- Discord: MrKirby#5063",
            "- Github: https://github.com/KSneijders/AoE2ScenarioParser/issues",
            "",
            "You can disable this check (and thereby this message) using:",
            ">>> from AoE2ScenarioParser import settings",
            ">>> settings.NOTIFY_UNKNOWN_BYTES = False",
            "",
            "Please be so kind and include the map in question. Thanks again!",
            "",
            f"Extra data found in the file ({len(retrieved_bytes)} bytes):",
            f"\t'{retrieved_bytes}'"
        ]))
        retriever.datatype.repeat = 1



from AoE2ScenarioParser.objects.aoe2_object import AoE2Object
from AoE2ScenarioParser.sections.retrievers.retriever_object_link import RetrieverObjectLink


class PlayerDataFour(AoE2Object):
    """Object for handling a tile in the map."""

    _link_list = [
        RetrieverObjectLink("population_limit", "Units", "player_data_4[__index__].population_limit"),
        RetrieverObjectLink("food_duplicate", "Units", "player_data_4[__index__].food_duplicate"),
        RetrieverObjectLink("wood_duplicate", "Units", "player_data_4[__index__].wood_duplicate"),
        RetrieverObjectLink("gold_duplicate", "Units", "player_data_4[__index__].gold_duplicate"),
        RetrieverObjectLink("stone_duplicate", "Units", "player_data_4[__index__].stone_duplicate"),
    ]

    def __init__(
            self,
            population_limit: float,
            food_duplicate: float,
            wood_duplicate: float,
            gold_duplicate: float,
            stone_duplicate: float,
            **kwargs):
        super().__init__(**kwargs)

        self.population_limit: float = population_limit
        self.food_duplicate: float = food_duplicate
        self.wood_duplicate: float = wood_duplicate
        self.gold_duplicate: float = gold_duplicate
        self.stone_duplicate: float = stone_duplicate



from uuid import UUID

from AoE2ScenarioParser.scenarios.support.data_triggers import DataTriggers


class ScenarioActions:
    def __init__(self, uuid: UUID) -> None:
        super().__init__()

        self._uuid = uuid

    def load_data_triggers(self, remove_template_triggers=True) -> DataTriggers:
        data_triggers = DataTriggers(self._uuid)
        data_triggers.discover(remove_template_triggers)

        return data_triggers



from __future__ import annotations

from enum import Enum
from typing import Union, List

from AoE2ScenarioParser.datasets.buildings import BuildingInfo
from AoE2ScenarioParser.datasets.trigger_lists import Age
from AoE2ScenarioParser.helper.list_functions import listify


class TechInfo(Enum):
    """
    This enum class provides information about most of the techs in the game. Information about the
    following properties of a tech is found in this class:
     - Tech ID
     - Icon ID

    **Methods**

    >>> TechInfo.from_id()
    >>> TechInfo.from_icon_id()
    >>> TechInfo.unique_techs()
    >>> TechInfo.unique_unit_upgrades()
    >>> TechInfo.town_center_techs()
    >>> TechInfo.blacksmith_techs()
    >>> TechInfo.monastery_techs()
    >>> TechInfo.university_techs()
    >>> TechInfo.eco_techs()

    **Examples**

    >>> TechInfo.LOOM.ID
    >>> 22

    >>> TechInfo.LOOM.ICON_ID
    >>> 6
    """

    @property
    def ID(self) -> int:
        """
        Returns:
            The ID of the specified tech
        """
        return self.value[0]

    @classmethod
    def from_id(cls, tech_id: int) -> TechInfo:
        """
        Get the TechInfo object from its ID

        Args:
            tech_id: The ID of the tech to get the TechInfo of

        Returns:
            A TechInfo object of the specified tech ID
        """
        if tech_id < 0:
            raise ValueError(f"{tech_id} is not a valid tech id value")
        for x in cls._member_map_.values():
            if x.value[0] == tech_id:
                return x

        raise KeyError(f"A technology with ID '{tech_id}' was not found in the dataset")

    @property
    def ICON_ID(self) -> int:
        """
        Returns:
            The icon ID of the specified tech
        """
        return self.value[1]

    @classmethod
    def from_icon_id(cls, tech_icon_id) -> TechInfo:
        """
        Get the TechInfo object from its icon ID

        Args:
            tech_icon_id: The icon ID of the tech to get the TechInfo of

        Returns:
            A TechInfo object of the tech with the specified icon ID
        """
        if tech_icon_id == -1:
            raise ValueError("-1 is not a valid icon_id value")
        for x in cls._member_map_.values():
            if x.value[1] == tech_icon_id:
                return x

        raise KeyError(f"A technology with icon id '{tech_icon_id}' was not found in the dataset")

    @staticmethod
    def unique_techs(exclude_castle_techs: bool = False, exclude_imp_techs: bool = False) -> List[TechInfo]:
        """
        Get the list of all the unique techs in the game

        Args:
            exclude_castle_techs: if set to True, exclude the castle age techs
            exclude_imp_techs: if set to True, exclude the imperial age techs

        Returns:
            A list of TechInfo objects which are all the unique techs in the game
        """
        unique_techs = {
            "castle_age": [
                TechInfo.ANARCHY,
                TechInfo.ANDEAN_SLING,
                TechInfo.ATLATL,
                TechInfo.BEARDED_AXE,
                TechInfo.CARRACK,
                TechInfo.CHATRAS,
                TechInfo.CHIEFTAINS,
                TechInfo.CORVINIAN_ARMY,
                TechInfo.DETINETS,
                TechInfo.FIRST_CRUSADE,
                TechInfo.GRAND_TRUNK_ROAD,
                TechInfo.GREAT_WALL,
                TechInfo.GREEK_FIRE,
                TechInfo.HILL_FORTS,
                TechInfo.HOWDAH,
                TechInfo.HULCHE_JAVELINEERS,
                TechInfo.INQUISITION,
                TechInfo.IRONCLAD,
                TechInfo.KAMANDARAN,
                TechInfo.KASBAH,
                TechInfo.KSHATRIYAS,
                TechInfo.NOMADS,
                TechInfo.MADRASAH,
                TechInfo.MARAUDERS,
                TechInfo.MEDICAL_CORPS,
                TechInfo.PAIKS,
                TechInfo.EUPSEONG,
                TechInfo.PAVISE,
                TechInfo.ROYAL_HEIRS,
                TechInfo.SHATAGNI,
                TechInfo.SILK_ARMOR,
                TechInfo.SIPAHI,
                TechInfo.STEPPE_HUSBANDRY,
                TechInfo.STIRRUPS,
                TechInfo.STRONGHOLD,
                TechInfo.SZLACHTA_PRIVILEGES,
                TechInfo.THALASSOCRACY,
                TechInfo.TIGUI,
                TechInfo.TUSK_SWORDS,
                TechInfo.WAGENBURG_TACTICS,
                TechInfo.YASAMA,
                TechInfo.YEOMEN,
                TechInfo.ZEALOTRY
            ],
            "imp_age": [
                TechInfo.ARQUEBUS,
                TechInfo.ARTILLERY,
                TechInfo.ATHEISM,
                TechInfo.BAGAINS,
                TechInfo.BERSERKERGANG,
                TechInfo.BURGUNDIAN_VINEYARDS,
                TechInfo.CHIVALRY,
                TechInfo.CRENELLATIONS,
                TechInfo.COUNTERWEIGHTS,
                TechInfo.FABRIC_SHIELDS,
                TechInfo.CUMAN_MERCENARIES,
                TechInfo.DOUBLE_CROSSBOW,
                TechInfo.DRILL,
                TechInfo.DRUZHINA,
                TechInfo.EL_DORADO,
                TechInfo.FARIMBA,
                TechInfo.FLEMISH_REVOLUTION,
                TechInfo.FORCED_LEVY,
                TechInfo.FRONTIER_GUARDS,
                TechInfo.FUROR_CELTICA,
                TechInfo.GARLAND_WARS,
                TechInfo.HUSSITE_REFORMS,
                TechInfo.KATAPARUTO,
                TechInfo.LECHITIC_LEGACY,
                TechInfo.LOGISTICA,
                TechInfo.MAGHREBI_CAMELS,
                TechInfo.MAHAYANA,
                TechInfo.MAHOUTS,
                TechInfo.MANIPUR_CAVALRY,
                TechInfo.PAPER_MONEY,
                TechInfo.PERFUSION,
                TechInfo.RECURVE_BOW,
                TechInfo.ROCKETRY,
                TechInfo.HAUBERK,
                TechInfo.SHINKICHON,
                TechInfo.SILK_ROAD,
                TechInfo.SUPREMACY,
                TechInfo.TIMURID_SIEGECRAFT,
                TechInfo.TORSION_ENGINES,
                TechInfo.TOWER_SHIELDS,
                TechInfo.WARWOLF,
                TechInfo.WOOTZ_STEEL,
            ]
        }

        techs_to_return = []

        if not exclude_castle_techs:
            techs_to_return.extend(unique_techs["castle_age"])
        if not exclude_imp_techs:
            techs_to_return.extend(unique_techs["imp_age"])

        return techs_to_return

    @staticmethod
    def unique_unit_upgrades(
            exclude_castle_techs: bool = False,
            exclude_non_castle_techs: bool = False
    ) -> List[TechInfo]:
        """
        Args:
            exclude_castle_techs: if set to false, excludes the castle unique unit techs from the list of techs returned
            exclude_non_castle_techs: if set to false, excludes the non castle unique unit techs from the list of techs returned

        Returns:
            A list of unique unite upgrade tech IDs
        """
        unique_techs = {
            "castle": [
                TechInfo.ELITE_ARAMBAI,
                TechInfo.ELITE_BALLISTA_ELEPHANT,
                TechInfo.ELITE_BERSERK,
                TechInfo.ELITE_BOYAR,
                TechInfo.ELITE_CAMEL_ARCHER,
                TechInfo.ELITE_CATAPHRACT,
                TechInfo.ELITE_CHAKRAM_THROWER,
                TechInfo.ELITE_CHU_KO_NU,
                TechInfo.ELITE_CONQUISTADOR,
                TechInfo.ELITE_COUSTILLIER,
                TechInfo.ELITE_GBETO,
                TechInfo.ELITE_GENOESE_CROSSBOWMAN,
                TechInfo.ELITE_GHULAM,
                TechInfo.ELITE_HUSKARL,
                TechInfo.ELITE_HUSSITE_WAGON,
                TechInfo.ELITE_JAGUAR_WARRIOR,
                TechInfo.ELITE_JANISSARY,
                TechInfo.ELITE_KAMAYUK,
                TechInfo.ELITE_KARAMBIT_WARRIOR,
                TechInfo.ELITE_KESHIK,
                TechInfo.ELITE_KIPCHAK,
                TechInfo.ELITE_KONNIK,
                TechInfo.ELITE_LEITIS,
                TechInfo.ELITE_LONGBOWMAN,
                TechInfo.ELITE_MAGYAR_HUSZAR,
                TechInfo.ELITE_MAMELUKE,
                TechInfo.ELITE_MANGUDAI,
                TechInfo.ELITE_OBUCH,
                TechInfo.ELITE_ORGAN_GUN,
                TechInfo.ELITE_PLUMED_ARCHER,
                TechInfo.ELITE_RATHA,
                TechInfo.ELITE_RATTAN_ARCHER,
                TechInfo.ELITE_SAMURAI,
                TechInfo.ELITE_SERJEANT,
                TechInfo.ELITE_SHOTEL_WARRIOR,
                TechInfo.ELITE_TARKAN,
                TechInfo.ELITE_TEUTONIC_KNIGHT,
                TechInfo.ELITE_THROWING_AXEMAN,
                TechInfo.ELITE_URUMI_SWORDSMAN,
                TechInfo.ELITE_WAR_ELEPHANT,
                TechInfo.ELITE_WAR_WAGON,
                TechInfo.ELITE_WOAD_RAIDER,
            ],
            "non_castle": [
                TechInfo.ELITE_CARAVEL,
                TechInfo.ELITE_GENITOUR,
                TechInfo.ELITE_LONGBOAT,
                TechInfo.ELITE_TURTLE_SHIP,
                TechInfo.HOUFNICE,
                TechInfo.IMPERIAL_CAMEL_RIDER,
                TechInfo.IMPERIAL_SKIRMISHER,
                TechInfo.ELITE_SHRIVAMSHA_RIDER,
                TechInfo.WINGED_HUSSAR
            ]
        }

        techs_to_return = []

        if not exclude_castle_techs:
            techs_to_return.extend(unique_techs["castle"])
        if not exclude_non_castle_techs:
            techs_to_return.extend(unique_techs["non_castle"])

        return techs_to_return

    @staticmethod
    def town_center_techs(ages: Union[int, List[int]] = None):
        """
        Args:
            ages: a list of age IDs (IDs are located in the Age IntEnum dataset). If specified, only techs from these
                ages are returned. If unspecified, all ages' techs are returned

        Returns:
            A list of TechInfo objects which are the blacksmith upgrade techs for the given age
        """
        ages = list(Age) if ages is None else listify(ages)

        upgrades = {
            Age.DARK_AGE: [
                TechInfo.LOOM,
                TechInfo.FEUDAL_AGE
            ],
            Age.FEUDAL_AGE: [
                TechInfo.WHEELBARROW,
                TechInfo.TOWN_WATCH,
                TechInfo.CASTLE_AGE
            ],
            Age.CASTLE_AGE: [
                TechInfo.HAND_CART,
                TechInfo.TOWN_PATROL,
                TechInfo.IMPERIAL_AGE
            ],
            Age.IMPERIAL_AGE: [],
        }

        techs_to_return = []
        for age in ages:
            techs_to_return.extend(upgrades[age])

        return techs_to_return

    @staticmethod
    def blacksmith_techs(ages: Union[int, List[int]] = None) -> List[TechInfo]:
        """
        Args:
            ages: a list of age IDs (IDs are located in the Age IntEnum dataset). If specified, only techs from these
                ages are returned. If unspecified, all ages' techs are returned

        Returns:
            A list of TechInfo objects which are the blacksmith upgrade techs for the given age
        """
        ages = list(Age) if ages is None else listify(ages)

        upgrades = {
            Age.DARK_AGE: [],
            Age.FEUDAL_AGE: [
                TechInfo.FORGING,
                TechInfo.SCALE_MAIL_ARMOR,
                TechInfo.SCALE_BARDING_ARMOR,
                TechInfo.FLETCHING,
                TechInfo.PADDED_ARCHER_ARMOR
            ],
            Age.CASTLE_AGE: [
                TechInfo.IRON_CASTING,
                TechInfo.CHAIN_MAIL_ARMOR,
                TechInfo.CHAIN_BARDING_ARMOR,
                TechInfo.BODKIN_ARROW,
                TechInfo.LEATHER_ARCHER_ARMOR
            ],
            Age.IMPERIAL_AGE: [
                TechInfo.BLAST_FURNACE,
                TechInfo.PLATE_MAIL_ARMOR,
                TechInfo.PLATE_BARDING_ARMOR,
                TechInfo.BRACER,
                TechInfo.RING_ARCHER_ARMOR
            ],
        }

        techs_to_return = []
        for age in ages:
            techs_to_return.extend(upgrades[age])

        return techs_to_return

    @staticmethod
    def monastery_techs(ages: Union[int, List[int]] = None) -> List[TechInfo]:
        """
        Args:
            ages: The age ID (IDs are located in the Age IntEnum dataset). If specified, only techs from these
                ages are returned. If unspecified, all ages' techs are returned

        Returns:
            A list of TechInfo objects which are the monastery upgrade techs for the given age
        """
        ages = list(Age) if ages is None else listify(ages)

        upgrades = {
            Age.DARK_AGE: [],
            Age.FEUDAL_AGE: [],
            Age.CASTLE_AGE: [
                TechInfo.REDEMPTION,
                TechInfo.ATONEMENT,
                TechInfo.HERBAL_MEDICINE,
                TechInfo.HERESY,
                TechInfo.SANCTITY,
                TechInfo.FERVOR,
            ],
            Age.IMPERIAL_AGE: [
                TechInfo.FAITH,
                TechInfo.ILLUMINATION,
                TechInfo.BLOCK_PRINTING,
                TechInfo.THEOCRACY
            ],
        }

        techs_to_return = []
        for age in ages:
            techs_to_return.extend(upgrades[age])

        return techs_to_return

    @staticmethod
    def university_techs(ages: Union[int, List[int]] = None) -> List[TechInfo]:
        """
        Args:
            ages: The age ID (IDs are located in the Age IntEnum dataset). If specified, only techs from these
                ages are returned. If unspecified, all ages' techs are returned

        Returns:
            A list of TechInfo objects which are the university upgrade techs for the given age
        """
        ages = list(Age) if ages is None else listify(ages)

        upgrades = {
            Age.DARK_AGE: [],
            Age.FEUDAL_AGE: [],
            Age.CASTLE_AGE: [
                TechInfo.MASONRY,
                TechInfo.FORTIFIED_WALL,
                TechInfo.BALLISTICS,
                TechInfo.GUARD_TOWER,
                TechInfo.HEATED_SHOT,
                TechInfo.MURDER_HOLES,
                TechInfo.TREADMILL_CRANE,
            ],
            Age.IMPERIAL_AGE: [
                TechInfo.ARCHITECTURE,
                TechInfo.CHEMISTRY,
                TechInfo.SIEGE_ENGINEERS,
                TechInfo.KEEP,
                TechInfo.ARROWSLITS,
                TechInfo.BOMBARD_TOWER
            ],
        }

        techs_to_return = []
        for age in ages:
            techs_to_return.extend(upgrades[age])

        return techs_to_return

    @staticmethod
    def eco_techs(ages: Union[int, List[int]] = None, buildings: Union[int, List[int]] = None) -> List[TechInfo]:
        """
        Args:
            ages: The age ID (IDs are located in the Age IntEnum dataset). If specified, only techs from these
                ages are returned. If unspecified, all ages' techs are returned
            buildings: The building ID for which the upgrades are needed. If unspecified, eco upgrades from all
                economic buildings (Mill, Mining Camp, Lumber Camp, Town Center and Market) are returned

        Returns:
            A list of TechInfo objects which are the university upgrade techs for the given age
        """
        ages = list(Age) if ages is None else listify(ages)

        if buildings is None:
            buildings = [
                BuildingInfo.MILL.ID,
                BuildingInfo.MINING_CAMP.ID,
                BuildingInfo.LUMBER_CAMP.ID,
                BuildingInfo.TOWN_CENTER.ID,
                BuildingInfo.MARKET.ID
            ]
        else:
            buildings = listify(buildings)

        upgrades = {
            Age.DARK_AGE: {
                BuildingInfo.MILL.ID: [],
                BuildingInfo.MINING_CAMP.ID: [],
                BuildingInfo.LUMBER_CAMP.ID: [],
                BuildingInfo.TOWN_CENTER.ID: [],
                BuildingInfo.MARKET.ID: [],
            },
            Age.FEUDAL_AGE: {
                BuildingInfo.MILL.ID: [TechInfo.HORSE_COLLAR],
                BuildingInfo.MINING_CAMP.ID: [TechInfo.GOLD_MINING, TechInfo.STONE_MINING],
                BuildingInfo.LUMBER_CAMP.ID: [TechInfo.DOUBLE_BIT_AXE],
                BuildingInfo.TOWN_CENTER.ID: [TechInfo.WHEELBARROW],
                BuildingInfo.MARKET.ID: [],
            },
            Age.CASTLE_AGE: {
                BuildingInfo.MILL.ID: [TechInfo.HEAVY_PLOW],
                BuildingInfo.MINING_CAMP.ID: [TechInfo.GOLD_SHAFT_MINING, TechInfo.STONE_SHAFT_MINING],
                BuildingInfo.LUMBER_CAMP.ID: [TechInfo.BOW_SAW],
                BuildingInfo.TOWN_CENTER.ID: [TechInfo.HAND_CART],
                BuildingInfo.MARKET.ID: [TechInfo.COINAGE, TechInfo.CARAVAN]
            },
            Age.IMPERIAL_AGE: {
                BuildingInfo.MILL.ID: [TechInfo.CROP_ROTATION],
                BuildingInfo.MINING_CAMP.ID: [],
                BuildingInfo.LUMBER_CAMP.ID: [TechInfo.TWO_MAN_SAW],
                BuildingInfo.TOWN_CENTER.ID: [],
                BuildingInfo.MARKET.ID: [TechInfo.BANKING, TechInfo.GUILDS]
            },
        }

        techs_to_return = []
        for age in ages:
            for building in buildings:
                techs_to_return.extend(upgrades[age][building])

        return techs_to_return

    ANARCHY = 16, 33
    ANDEAN_SLING = 516, 33
    ARBALESTER = 237, 54
    ARCHITECTURE = 51, 14
    ARQUEBUS = 573, 107
    ARROWSLITS = 608, 119
    ARSON = 602, 118
    ARTILLERY = 10, 107
    ATHEISM = 21, 107
    ATLATL = 460, 33
    ATONEMENT = 319, 93
    AZTECS = 543, -1
    BAGAINS = 686, 107
    BALLISTICS = 93, 25
    BANKING = 17, 3
    BEARDED_AXE = 83, 107
    BERBERS = 583, -1
    BERSERKERGANG = 49, 107
    BLAST_FURNACE = 75, 21
    BLOCK_PRINTING = 230, 82
    BLOODLINES = 435, 110
    BODKIN_ARROW = 200, 35
    BOMBARD_CANNON = 188, -1
    BOMBARD_TOWER = 64, 47
    BOW_SAW = 203, 71
    BRACER = 201, 37
    BRITONS = 529, -1
    BULGARIANS = 673, -1
    BURMESE = 652, -1
    BYZANTINES = 535, -1
    CANNON_GALLEON = 37, 9
    CAPPED_RAM = 96, 27
    CARAVAN = 48, 113
    CAREENING = 374, 98
    CARRACK = 572, 33
    CARTOGRAPHY = 19, 4
    CASTLE_AGE = 102, 31
    CAVALIER = 209, 78
    CELTS = 541, -1
    CHAIN_BARDING_ARMOR = 82, 23
    CHAIN_MAIL_ARMOR = 76, 22
    CHAMPION = 264, 44
    CHATRAS = 628, 33
    CHEMISTRY = 47, 12
    CHINESE = 534, -1
    CHIVALRY = 493, 33
    COINAGE = 23, 7
    CONSCRIPTION = 315, 91
    CORVINIAN_ARMY = 514, 33
    FABRIC_SHIELDS = 517, 107
    CRENELLATIONS = 11, 107
    CROP_ROTATION = 12, 0
    CROSSBOWMAN = 100, 29
    CHINESE_50_PERCENT_HP_DEMOS = 396, -1
    CUMAN_MERCENARIES = 690, 107
    CUMANS = 675, -1
    DARK_AGE = 104, -1
    DOUBLE_CROSSBOW = 623, 107
    FRANKS_FREE_FARMING_1 = 287, -1
    DOUBLE_BIT_AXE = 202, 70
    DRILL = 6, 107
    DRUZHINA = 513, 107
    DRY_DOCK = 375, 99
    EAGLE_WARRIOR = 384, 75
    EL_DORADO = 4, 107
    ELITE_ARAMBAI = 619, 105
    ELITE_BALLISTA_ELEPHANT = 615, 105
    ELITE_BATTLE_ELEPHANT = 631, 121
    ELITE_BERSERK = 398, 105
    ELITE_BOYAR = 504, 105
    ELITE_CAMEL_ARCHER = 565, 105
    ELITE_CANNON_GALLEON = 376, 100
    ELITE_CARAVEL = 597, 105
    ELITE_CATAPHRACT = 361, 105
    ELITE_COUSTILLIER = 751, 105
    ELITE_CHU_KO_NU = 362, 105
    ELITE_CONQUISTADOR = 60, 105
    ELITE_EAGLE_WARRIOR = 434, 115
    ELITE_ELEPHANT_ARCHER = 481, 105
    ELITE_GBETO = 567, 105
    ELITE_GENITOUR = 599, 105
    ELITE_GENOESE_CROSSBOWMAN = 468, 105
    ELITE_HUSKARL = 365, 105
    ELITE_JAGUAR_WARRIOR = 432, 105
    ELITE_JANISSARY = 369, 105
    ELITE_KAMAYUK = 509, 105
    ELITE_KARAMBIT_WARRIOR = 617, 105
    ELITE_KESHIK = 680, 105
    ELITE_KIPCHAK = 682, 105
    CHIEFTAINS = 463, 33
    ELITE_KONNIK = 678, 105
    ELITE_LEITIS = 684, 105
    ELITE_LONGBOAT = 372, 105
    ELITE_LONGBOWMAN = 360, 105
    ELITE_MAGYAR_HUSZAR = 472, 105
    ELITE_MAMELUKE = 368, 105
    ELITE_MANGUDAI = 371, 105
    ELITE_ORGAN_GUN = 563, 105
    ELITE_PLUMED_ARCHER = 27, 105
    ELITE_RATTAN_ARCHER = 621, 105
    ELITE_SAMURAI = 366, 105
    ELITE_SERJEANT = 753, 105
    ELITE_SHOTEL_WARRIOR = 569, 105
    ELITE_SKIRMISHER = 98, 28
    ELITE_STEPPE_LANCER = 715, 123
    ELITE_TARKAN = 2, 105
    ELITE_TEUTONIC_KNIGHT = 364, 105
    ELITE_THROWING_AXEMAN = 363, 105
    ELITE_TURTLE_SHIP = 448, 105
    ELITE_WAR_ELEPHANT = 367, 105
    ELITE_WAR_WAGON = 450, 105
    ELITE_WOAD_RAIDER = 370, 105
    ENABLE_COWS = 557, 94
    ENABLE_LLAMAS = 556, 94
    ENABLE_SHEEP = 555, 94
    ENABLE_TURKEYS = 558, 94
    ETHIOPIANS = 581, -1
    FAITH = 45, 11
    FARIMBA = 577, 107
    FAST_FIRE_SHIP = 246, 40
    FERVOR = 252, 73
    FEUDAL_AGE = 101, 30
    FIRE_TOWER = 527, -1
    FLETCHING = 199, 34
    FORCED_LEVY = 625, 107
    FORGING = 67, 17
    FORTIFIED_WALL = 194, 46
    FRANKS = 530, -1
    FREE_CARTOGRAPHY = 600, -1
    FUROR_CELTICA = 5, 107
    GALLEON = 35, 59
    GARLAND_WARS = 24, 107
    GILLNETS = 65, 41
    GOLD_MINING = 55, 15
    GOLD_SHAFT_MINING = 182, 62
    GOTHS = 531, -1
    GREAT_WALL = 462, 33
    GREEK_FIRE = 464, 33
    GUARD_TOWER = 140, 76
    GUILDS = 15, 58
    HALBERDIER = 429, 106
    HAND_CANNON = 85, -1
    HAND_CART = 249, 42
    HEATED_SHOT = 380, 104
    HEAVY_CAMEL_RIDER = 236, 55
    HEAVY_CAVALRY_ARCHER = 218, 52
    HEAVY_DEMOLITION_SHIP = 244, 39
    HEAVY_PLOW = 13, 1
    HEAVY_SCORPION = 239, 38
    HERBAL_MEDICINE = 441, 114
    HERESY = 439, 108
    HILL_FORTS_2 = 394, -1
    HILL_FORTS = 691, 33
    HOARDINGS = 379, 103
    HORSE_COLLAR = 14, 2
    HOWDAH = 626, 33
    HUNS = 545, -1
    HUNTING_DOGS = 526, 94
    HUSBANDRY = 39, 10
    HUSSAR = 428, 122
    ILLUMINATION = 233, 84
    IMPERIAL_AGE = 103, 32
    IMPERIAL_CAMEL_RIDER = 521, 74
    IMPERIAL_SKIRMISHER = 655, 120
    INCAS = 549, -1
    HINDUSTANIS = 548, -1
    INQUISITION = 492, 33
    IRON_CASTING = 68, 18
    IRONCLAD = 489, 33
    ITALIANS = 547, -1
    JAPANESE = 533, -1
    KAMANDARAN = 488, 33
    KASBAH = 578, 33
    KATAPARUTO = 59, 107
    KEEP = 63, 16
    KHMER = 650, -1
    KOREANS = 546, -1
    LEATHER_ARCHER_ARMOR = 212, 50
    LIGHT_CAVALRY = 254, 43
    LITHUANIANS = 676, -1
    LOGISTICA = 61, 107
    LONG_SWORDSMAN = 207, 48
    LOOM = 22, 6
    MADRASAH = 490, 33
    MAGHREBI_CAMELS = 579, 107
    MAGYARS = 550, -1
    MAHOUTS = 7, 107
    MALAY = 651, -1
    MALIANS = 582, -1
    MAN_AT_ARMS = 222, 85
    MANIPUR_CAVALRY = 627, 107
    MARAUDERS = 483, 33
    MASONRY = 50, 13
    MAYANS = 544, -1
    MONGOLS = 540, -1
    MURDER_HOLES = 322, 61
    NOMADS = 487, 33
    HULCHE_JAVELINEERS = 485, 33
    ONAGER = 257, 57
    ORTHODOXY = 512, 33
    PADDED_ARCHER_ARMOR = 211, 49
    PALADIN = 265, 45
    EUPSEONG = 486, 33
    PAPER_MONEY = 629, 107
    PARTHIAN_TACTICS = 436, 111
    PAVISE = 494, 33
    PERFUSION = 457, 107
    PERSIANS = 536, -1
    PIKEMAN = 197, 36
    PLATE_BARDING_ARMOR = 80, 65
    PLATE_MAIL_ARMOR = 77, 64
    PORTUGUESE = 580, -1
    RECURVE_BOW = 515, 107
    REDEMPTION = 316, 92
    REVETMENTS = 525, -1
    RING_ARCHER_ARMOR = 219, 51
    ROCKETRY = 52, 107
    ROYAL_HEIRS = 574, 33
    SANCTITY = 231, 83
    SAPPERS = 321, 5
    SARACENS = 537, -1
    SCALE_BARDING_ARMOR = 81, 66
    SCALE_MAIL_ARMOR = 74, 63
    SCORPION = 94, -1
    SET_MAXIMUM_POPULATION_NO_HOUSES = 658, -1
    SHATAGNI = 507, 107
    SHINKICHON = 445, 107
    SHIPWRIGHT = 373, 97
    SIEGE_ENGINEERS = 377, 101
    SIEGE_ONAGER = 320, 96
    SIEGE_RAM = 255, 86
    SILK_ARMOR = 687, 33
    SILK_ROAD = 499, 107
    SIPAHI = 491, 33
    SLAVS = 551, -1
    SPANISH = 542, -1
    SPIES_AND_TREASON = 408, 19
    SQUIRES = 215, 80
    STEPPE_HUSBANDRY = 689, 33
    STIRRUPS = 685, 33
    STONE_MINING = 278, 87
    STONE_SHAFT_MINING = 279, 88
    STRONGHOLD = 482, 33
    GRAND_TRUNK_ROAD = 506, 33
    SUPPLIES = 716, 124
    SUPREMACY = 440, 107
    TATARS = 674, -1
    TEUTONS = 532, -1
    THALASSOCRACY = 624, 33
    THEOCRACY = 438, 109
    THUMB_RING = 437, 112
    TIGUI = 576, 33
    TIMURID_SIEGECRAFT = 688, 107
    TORSION_ENGINES = 575, 107
    TOWER_SHIELDS = 692, 107
    TOWN_CENTER_SPAWN = 639, -1
    TOWN_PATROL = 280, 89
    TOWN_WATCH = 8, 69
    TRACKING = 90, 67
    TREADMILL_CRANE = 54, 60
    TURKS = 538, -1
    TUSK_SWORDS = 622, 33
    TWO_HANDED_SWORDSMAN = 217, 53
    TWO_MAN_SAW = 221, 81
    VIETNAMESE = 653, -1
    VIETNAMESE_VISION = 665, -1
    VIKINGS = 539, -1
    WAR_GALLEY = 34, 8
    WARWOLF = 461, 107
    WHEELBARROW = 213, 79
    YASAMA = 484, 33
    YEOMEN = 3, 33
    ZEALOTRY = 9, 107
    SICILIANS = 749, -1
    BURGUNDIANS = 748, -1
    BURGUNDIAN_VINEYARDS = 754, 33
    FLEMISH_REVOLUTION = 755, 107
    FIRST_CRUSADE = 756, 33
    HAUBERK = 757, 107
    CONVERSION_ENABLER = 243, -1
    POLES = 776, -1
    SZLACHTA_PRIVILEGES = 782, 33
    LECHITIC_LEGACY = 783, 107
    FOLWARK = 793, -1
    FOLWARK_FEUDAL = 794, -1
    FOLWARK_CASTLE = 795, -1
    FOLWARK_IMPERIAL = 796, -1
    HORSE_COLLAR_FOLWARK_BONUS_INCREASE = 797, -1
    FOLWARK_HEAVY_PLOW_EXTRA = 797, -1
    FOLWARK_CROP_ROTATION_EXTRA = 797, -1
    STONE_MINING_GOLD_GENERATION_INCREASE = 806, -1
    STONE_SHAFT_MINING_GOLD_GENERATION_INCREASE = 807, -1
    BOHEMIANS = 777, -1
    WAGENBURG_TACTICS = 784, 33
    HUSSITE_REFORMS = 785, 107
    WINGED_HUSSAR = 786, 125
    HOUFNICE = 787, 126
    ELITE_HUSSITE_WAGON = 781, 105
    ELITE_OBUCH = 779, 105
    SET_MAX_POP = 658, -1
    ENABLE_MONUMENT_RES_TRICKLE = 729, -1
    RESOURCES_LAST_15_PERCENT_LONGER = 737, -1
    RESOURCES_LAST_30_PERCENT_LONGER = 738, -1
    RESOURCES_LAST_40_PERCENT_LONGER = 739, -1
    RESOURCES_LAST_50_PERCENT_LONGER = 740, -1
    RESOURCES_LAST_75_PERCENT_LONGER = 741, -1
    RESOURCES_LAST_100_PERCENT_LONGER = 742, -1
    RESOURCES_LAST_125_PERCENT_LONGER = 743, -1
    RESOURCES_LAST_150_PERCENT_LONGER = 744, -1
    RESOURCES_LAST_175_PERCENT_LONGER = 745, -1
    RESOURCES_LAST_200_PERCENT_LONGER = 746, -1
    RESOURCES_LAST_300_PERCENT_LONGER = 747, -1
    DISABLE_FREE_TRANSPORT = 229, -1
    AUTO_UPGRADE_SCOUT_FEUDAL_AGE = 20, 0
    COMPASS = 28, 827
    SPANISH_CANNON_GALLEON = 57, 0
    WALLS_HP_CASTLE_AGE = 71, -1
    PALISADE_WALLS_HP_FEUDAL_AGE = 72, -1
    FREE_GUARD_TOWER = 442, -1
    FREE_KEEP = 443, -1
    FREE_BOMBARD_TOWER = 444, -1
    COUNTERWEIGHTS = 454, 107
    DETINETS = 455, 33
    ARROWSLITS_GUARD_TOWER = 610, -1
    ARROWSLITS_KEEP = 611, -1
    FIRE_SHIPS_AUTO_BALLISTICS = 612, -1
    XOLOTL_POST_IMPERIAL_UPGRADE = 636, -1
    IMPERIAL_NOMADS = 641, -1
    THALASSOCRACY_AND_FLETCHING = 669, -1
    THALASSOCRACY_AND_BODKIN = 670, -1
    THALASSOCRACY_AND_BRACER = 671, -1
    RENAME_UNITS = 713, -1
    FTT_MOVE_LANCERS = 717, 0
    FTT_MOVE_TARKANS = 718, 0
    FTT_MOVE_HUSKARLS = 719, 0
    FTT_MOVE_SLINGERS = 720, 0
    CHANGE_TC_CONSTR_TIME1 = 723, 0
    CHANGE_TC_CONSTR_TIME2 = 724, 0
    THALASSOCRACY_AND_CHEMISTRY = 734, -1
    THALASSOCRACY_AND_MURDER_HOLES = 735, -1
    THALASSOCRACY_AND_HEATED_SHOT = 736, -1
    FEUDAL_ECO_TECH_REQUIREMENT = 758, -1
    CASTLE_ECO_TECH_REQUIREMENT = 759, -1
    IMPERIAL_ECO_TECH_REQUIREMENT = 760, -1
    HEAVY_PLOW_REQUIREMENT = 761, -1
    BOW_SAW_REQUIREMENT = 762, -1
    HAND_CART_REQUIREMENT = 763, -1
    GOLD_SHAFT_MINING_REQUIREMENT = 764, -1
    STONE_SHAFT_MINING_REQUIREMENT = 765, -1
    CROP_ROTATION_REQUIREMENT = 766, -1
    TWO_MAN_SAW_REQUIREMENT = 767, -1
    CAVALIER_REQUIREMENT = 768, -1
    WINGED_HUSSAR_LITHUANIANS = 788, -1
    WINGED_HUSSAR_POLES = 789, -1
    HEAVY_PLOW_FOLWARK_BONUS_INCREASE = 798, -1
    CROP_ROTATION_FOLWARK_BONUS_INCREASE = 799, -1
    HORSE_COLLAR_AND_CHINESE_TB = 812, -1
    HEAVY_PLOW_AND_CHINESE_TB = 813, -1
    CROP_ROTATION_AND_CHINESE_TB = 814, -1
    HORSE_COLLAR_AND_CHINESE_TB_AND_SICILIANS = 815, -1
    HEAVY_PLOW_AND_CHINESE_TB_AND_SICILIANS = 816, -1
    CROP_ROTATION_AND_CHINESE_TB_AND_SICILIANS = 817, -1
    CHINESE_TB_AND_POLES = 818, -1
    HORSE_COLLAR_AND_CHINESE_TB_AND_POLES = 819, -1
    HEAVY_PLOW_AND_CHINESE_TB_AND_POLES = 820, -1
    CROP_ROTATION_AND_CHINESE_TB_AND_POLES = 821, -1
    DRAVIDIANS = 822, -1
    BENGALIS = 823, -1
    GURJARAS = 824, -1
    ELITE_URUMI_SWORDSMAN = 826, 105
    ELITE_RATHA = 828, 105
    ELITE_CHAKRAM_THROWER = 830, 105
    MEDICAL_CORPS = 831, 33
    WOOTZ_STEEL = 832, 107
    PAIKS = 833, 33
    MAHAYANA = 834, 107
    KSHATRIYAS = 835, 33
    FRONTIER_GUARDS = 836, 107
    SIEGE_ELEPHANT = 838, 128
    ELITE_GHULAM = 840, 105
    ELITE_SHRIVAMSHA_RIDER = 843, 105
    CAMEL_SCOUT_EXTRA_BONUS_DAMAGE = 859, -1
    UPGRADE_CAMEL_SCOUTS_TO_RIDERS = 860, -1
    KSHATRIYAS_AND_SUPPLIES = 861, -1
    FTT_MOVE_TARKANS_2 = 862, 0
    FTT_MOVE_ELEPHANT_ARCHERS = 863, 0
    FTT_MOVE_ARMORED_ELEPHANTS = 864, 0
    FTT_MOVE_SHRIVAMSHA_RIDERS = 865, 0
    FTT_MOVE_SHRIVAMSHA_RIDERS_2 = 866, 0
    PAPER_MONEY_AND_DOUBLE_BIT_AXE = 867, -1
    PAPER_MONEY_AND_BOW_SAW = 868, -1
    PAPER_MONEY_AND_TWO_MAN_SAW = 869, -1




from __future__ import annotations

from typing import List, Tuple, Union, Any

from AoE2ScenarioParser.datasets import effects
from AoE2ScenarioParser.datasets.effects import EffectId
from AoE2ScenarioParser.datasets.players import PlayerColorId, PlayerId
from AoE2ScenarioParser.datasets.trigger_lists import ObjectAttribute
from AoE2ScenarioParser.helper.helper import raise_if_not_int_subclass, value_is_valid, validate_coords
from AoE2ScenarioParser.helper.printers import warn
from AoE2ScenarioParser.helper.string_manipulations import add_tabs
from AoE2ScenarioParser.objects.aoe2_object import AoE2Object
from AoE2ScenarioParser.objects.support.attr_presentation import transform_effect_attr_value
from AoE2ScenarioParser.objects.support.trigger_object import TriggerComponent
from AoE2ScenarioParser.scenarios.scenario_store import getters
from AoE2ScenarioParser.sections.retrievers.retriever_object_link import RetrieverObjectLink
from AoE2ScenarioParser.sections.retrievers.support import Support


def _add_trail_if_string_attr_is_used_in_effect(obj: Effect, attr_name, val: Union[bytes, str]):
    if attr_name in effects.attributes[obj.effect_type]:
        return val + (b"\x00" if type(val) is bytes else "\x00")
    return val


class Effect(AoE2Object, TriggerComponent):
    """Object for handling an effect."""
    hidden_attribute = 'effect_type'

    _link_list = [
        RetrieverObjectLink("effect_type", "Triggers", "trigger_data[__index__].effect_data[__index__].effect_type"),
        RetrieverObjectLink("ai_script_goal", "Triggers",
                            "trigger_data[__index__].effect_data[__index__].ai_script_goal"),
        RetrieverObjectLink("quantity", "Triggers", "trigger_data[__index__].effect_data[__index__].quantity"),
        RetrieverObjectLink("tribute_list", "Triggers", "trigger_data[__index__].effect_data[__index__].tribute_list"),
        RetrieverObjectLink("diplomacy", "Triggers", "trigger_data[__index__].effect_data[__index__].diplomacy"),
        RetrieverObjectLink("legacy_location_object_reference", "Triggers",
                            "trigger_data[__index__].effect_data[__index__].legacy_location_object_reference"),
        RetrieverObjectLink("object_list_unit_id", "Triggers",
                            "trigger_data[__index__].effect_data[__index__].object_list_unit_id"),
        RetrieverObjectLink("source_player", "Triggers",
                            "trigger_data[__index__].effect_data[__index__].source_player"),
        RetrieverObjectLink("target_player", "Triggers",
                            "trigger_data[__index__].effect_data[__index__].target_player"),
        RetrieverObjectLink("technology", "Triggers", "trigger_data[__index__].effect_data[__index__].technology"),
        RetrieverObjectLink("string_id", "Triggers", "trigger_data[__index__].effect_data[__index__].string_id"),
        RetrieverObjectLink("display_time", "Triggers", "trigger_data[__index__].effect_data[__index__].display_time"),
        RetrieverObjectLink("trigger_id", "Triggers", "trigger_data[__index__].effect_data[__index__].trigger_id"),
        RetrieverObjectLink("location_x", "Triggers", "trigger_data[__index__].effect_data[__index__].location_x"),
        RetrieverObjectLink("location_y", "Triggers", "trigger_data[__index__].effect_data[__index__].location_y"),
        RetrieverObjectLink("location_object_reference", "Triggers",
                            "trigger_data[__index__].effect_data[__index__].location_object_reference"),
        RetrieverObjectLink("area_x1", "Triggers", "trigger_data[__index__].effect_data[__index__].area_x1"),
        RetrieverObjectLink("area_y1", "Triggers", "trigger_data[__index__].effect_data[__index__].area_y1"),
        RetrieverObjectLink("area_x2", "Triggers", "trigger_data[__index__].effect_data[__index__].area_x2"),
        RetrieverObjectLink("area_y2", "Triggers", "trigger_data[__index__].effect_data[__index__].area_y2"),
        RetrieverObjectLink("object_group", "Triggers", "trigger_data[__index__].effect_data[__index__].object_group"),
        RetrieverObjectLink("object_type", "Triggers", "trigger_data[__index__].effect_data[__index__].object_type"),
        RetrieverObjectLink("instruction_panel_position", "Triggers",
                            "trigger_data[__index__].effect_data[__index__].instruction_panel_position"),
        RetrieverObjectLink("attack_stance", "Triggers",
                            "trigger_data[__index__].effect_data[__index__].attack_stance"),
        RetrieverObjectLink("time_unit", "Triggers", "trigger_data[__index__].effect_data[__index__].time_unit"),
        RetrieverObjectLink("enabled", "Triggers",
                            "trigger_data[__index__].effect_data[__index__].enabled"),
        RetrieverObjectLink("food", "Triggers", "trigger_data[__index__].effect_data[__index__].food"),
        RetrieverObjectLink("wood", "Triggers", "trigger_data[__index__].effect_data[__index__].wood"),
        RetrieverObjectLink("stone", "Triggers", "trigger_data[__index__].effect_data[__index__].stone"),
        RetrieverObjectLink("gold", "Triggers", "trigger_data[__index__].effect_data[__index__].gold"),
        RetrieverObjectLink("item_id", "Triggers", "trigger_data[__index__].effect_data[__index__].item_id"),
        RetrieverObjectLink("flash_object", "Triggers", "trigger_data[__index__].effect_data[__index__].flash_object"),
        RetrieverObjectLink("force_research_technology", "Triggers",
                            "trigger_data[__index__].effect_data[__index__].force_research_technology"),
        RetrieverObjectLink("visibility_state", "Triggers",
                            "trigger_data[__index__].effect_data[__index__].visibility_state"),
        RetrieverObjectLink("scroll", "Triggers", "trigger_data[__index__].effect_data[__index__].scroll"),
        RetrieverObjectLink("operation", "Triggers", "trigger_data[__index__].effect_data[__index__].operation"),
        RetrieverObjectLink("object_list_unit_id_2", "Triggers",
                            "trigger_data[__index__].effect_data[__index__].object_list_unit_id_2"),
        RetrieverObjectLink("button_location", "Triggers",
                            "trigger_data[__index__].effect_data[__index__].button_location"),
        RetrieverObjectLink("ai_signal_value", "Triggers",
                            "trigger_data[__index__].effect_data[__index__].ai_signal_value"),
        RetrieverObjectLink("object_attributes", "Triggers",
                            "trigger_data[__index__].effect_data[__index__].object_attributes"),
        RetrieverObjectLink("variable", "Triggers",
                            "trigger_data[__index__].effect_data[__index__].variable"),
        RetrieverObjectLink("timer", "Triggers",
                            "trigger_data[__index__].effect_data[__index__].timer"),
        RetrieverObjectLink("facet", "Triggers", "trigger_data[__index__].effect_data[__index__].facet"),
        RetrieverObjectLink("play_sound", "Triggers", "trigger_data[__index__].effect_data[__index__].play_sound"),
        RetrieverObjectLink("player_color", "Triggers", "trigger_data[__index__].effect_data[__index__].player_color",
                            Support(since=1.40)),
        RetrieverObjectLink("color_mood", "Triggers", "trigger_data[__index__].effect_data[__index__].color_mood",
                            Support(since=1.42)),
        RetrieverObjectLink("reset_timer", "Triggers", "trigger_data[__index__].effect_data[__index__].reset_timer",
                            Support(since=1.44)),
        RetrieverObjectLink("object_state", "Triggers", "trigger_data[__index__].effect_data[__index__].object_state",
                            Support(since=1.44)),
        RetrieverObjectLink("action_type", "Triggers", "trigger_data[__index__].effect_data[__index__].action_type",
                            Support(since=1.44)),
        RetrieverObjectLink("message", "Triggers", "trigger_data[__index__].effect_data[__index__].message",
                            commit_callback=_add_trail_if_string_attr_is_used_in_effect),
        RetrieverObjectLink("sound_name", "Triggers", "trigger_data[__index__].effect_data[__index__].sound_name",
                            commit_callback=_add_trail_if_string_attr_is_used_in_effect),
        RetrieverObjectLink("selected_object_ids", "Triggers",
                            "trigger_data[__index__].effect_data[__index__].selected_object_ids"),
    ]

    def __init__(self,
                 effect_type: int = None,
                 ai_script_goal: int = None,
                 armour_attack_quantity: int = None,
                 armour_attack_class: int = None,
                 quantity: int = None,
                 tribute_list: int = None,
                 diplomacy: int = None,
                 legacy_location_object_reference: int = None,
                 object_list_unit_id: int = None,
                 source_player: int = None,
                 target_player: int = None,
                 technology: int = None,
                 string_id: int = None,
                 display_time: int = None,
                 trigger_id: int = None,
                 location_x: int = None,
                 location_y: int = None,
                 location_object_reference: int = None,
                 area_x1: int = None,
                 area_y1: int = None,
                 area_x2: int = None,
                 area_y2: int = None,
                 object_group: int = None,
                 object_type: int = None,
                 instruction_panel_position: int = None,
                 attack_stance: int = None,
                 time_unit: int = None,
                 enabled: int = None,
                 food: int = None,
                 wood: int = None,
                 stone: int = None,
                 gold: int = None,
                 item_id: int = None,  # Unused (?)
                 flash_object: int = None,
                 force_research_technology: int = None,
                 visibility_state: int = None,
                 scroll: int = None,
                 operation: int = None,
                 object_list_unit_id_2: int = None,
                 button_location: int = None,
                 ai_signal_value: int = None,
                 object_attributes: int = None,
                 variable: int = None,
                 timer: int = None,
                 facet: int = None,
                 play_sound: int = None,
                 player_color: int = None,
                 color_mood: int = None,
                 reset_timer: int = None,
                 object_state: int = None,
                 action_type: int = None,
                 message: str = None,
                 sound_name: str = None,
                 selected_object_ids: List[int] = None,
                 **kwargs
                 ):
        super().__init__(**kwargs)

        raise_if_not_int_subclass([object_list_unit_id, technology, object_list_unit_id_2])

        if selected_object_ids is None:
            selected_object_ids = []

        # Set flags
        self._armour_attack_flag = _set_armour_attack_flag(effect_type, object_attributes)

        # Handle armour effect attributes:
        #   When effect is created through trigger.new_effect, aa values will be -1.
        #   If created while reading a scenario, they both default to None.
        if self._armour_attack_flag:
            # If effect created through reading scenario file
            if quantity is not None and armour_attack_class is None and armour_attack_quantity is None:
                armour_attack_class, armour_attack_quantity = self._quantity_to_aa(quantity)
                quantity = None
            # If effect created through new_effect with aa values defined
            elif value_is_valid(armour_attack_class) or value_is_valid(armour_attack_quantity):
                quantity = None
            # If created through new_effect with quantity defined instead of the aa values. Handled by quantity property
            else:
                pass
        else:
            armour_attack_class = armour_attack_quantity = None

        area_x1, area_y1, area_x2, area_y2 = validate_coords(area_x1, area_y1, area_x2, area_y2)

        if value_is_valid(legacy_location_object_reference):
            location_object_reference = legacy_location_object_reference

        # Bypass the @property which causes: self._update_armour_attack_flag()
        self._effect_type: int = effect_type
        self.ai_script_goal: int = ai_script_goal
        self.armour_attack_quantity: int = armour_attack_quantity
        self.armour_attack_class: int = armour_attack_class
        self.quantity: int = quantity
        self.tribute_list: int = tribute_list
        self.diplomacy: int = diplomacy
        self.object_list_unit_id: int = object_list_unit_id
        self.source_player: int = source_player
        self.target_player: int = target_player
        self.technology: int = technology
        self.string_id: int = string_id
        self.display_time: int = display_time
        self.trigger_id: int = trigger_id
        self.location_x: int = location_x
        self.location_y: int = location_y
        self.location_object_reference: int = location_object_reference
        self.area_x1: int = area_x1
        self.area_y1: int = area_y1
        self.area_x2: int = area_x2
        self.area_y2: int = area_y2
        self.object_group: int = object_group
        self.object_type: int = object_type
        self.instruction_panel_position: int = instruction_panel_position
        self.attack_stance: int = attack_stance
        self.time_unit: int = time_unit
        self.enabled: int = enabled
        self.food: int = food
        self.wood: int = wood
        self.stone: int = stone
        self.gold: int = gold
        # self.item_id: int = item_id  # Unused (?)
        self.flash_object: int = flash_object
        self.force_research_technology: int = force_research_technology
        self.visibility_state: int = visibility_state
        self.scroll: int = scroll
        self.operation: int = operation
        self.object_list_unit_id_2: int = object_list_unit_id_2
        self.button_location: int = button_location
        self.ai_signal_value: int = ai_signal_value
        self.object_attributes: int = object_attributes
        self.variable: int = variable
        self.timer: int = timer
        self.facet: int = facet
        self.play_sound: int = play_sound
        self.player_color: int = player_color
        self.color_mood: int = color_mood
        self.reset_timer = reset_timer
        self.object_state = object_state
        self.action_type = action_type
        self.message: str = message
        self.sound_name: str = sound_name
        self.selected_object_ids: List[int] = selected_object_ids

    @property
    def legacy_location_object_reference(self) -> int:
        return -1

    @property
    def player_color(self):
        return self._player_color

    @player_color.setter
    def player_color(self, value):
        if type(value) in [PlayerColorId, PlayerId]:
            value -= 1
        self._player_color = value

    @property
    def item_id(self):
        return self.object_list_unit_id

    @item_id.setter
    def item_id(self, value):
        raise ValueError("The `item_id` attribute isn't used in scenario's. Please use `object_list_unit_id`.\n"
                         "If you believe this is an error, please create a report at GitHub.")

    @property
    def effect_type(self):
        return self._effect_type

    @effect_type.setter
    def effect_type(self, value):
        self._effect_type = value
        self._update_armour_attack_flag()

    @property
    def object_attributes(self):
        return self._object_attributes

    @object_attributes.setter
    def object_attributes(self, value):
        self._object_attributes = value
        self._update_armour_attack_flag()

    @property
    def armour_attack_quantity(self):
        return self._armour_attack_quantity

    @armour_attack_quantity.setter
    def armour_attack_quantity(self, value):
        if value is not None and value != [] and not self._armour_attack_flag:
            warn("Setting 'effect.armour_attack_quantity' when the effect doesn't use armour/attack attributes "
                 "might result in unintended behaviour.")
        self._armour_attack_quantity = value

    @property
    def armour_attack_class(self):
        return self._armour_attack_class

    @armour_attack_class.setter
    def armour_attack_class(self, value):
        if value is not None and value != [] and not self._armour_attack_flag:
            warn("Setting 'effect.armour_attack_class' when the effect doesn't use armour/attack attributes "
                 "might result in unintended behaviour.")
        self._armour_attack_class = value

    @property
    def quantity(self):
        if self._armour_attack_flag:
            return self._aa_to_quantity(self.armour_attack_quantity, self.armour_attack_class)
        return self._quantity

    @quantity.setter
    def quantity(self, value):
        # Quantity by default, when unused is []
        if value is not None and value != [] and self._armour_attack_flag:
            warn("Setting 'effect.quantity' directly in an effect that uses armour/attack attributes "
                 "might result in unintended behaviour.\nPlease use the 'effect.armour_attack_quantity' "
                 "and 'effect.armour_attack_class' attributes instead.")
            self.armour_attack_class, self.armour_attack_quantity = self._quantity_to_aa(value)
        self._quantity = value

    @property
    def selected_object_ids(self) -> List[int]:
        return self._selected_object_ids

    @selected_object_ids.setter
    def selected_object_ids(self, val: List[int]):
        if type(val) is int:
            val = [val]
        self._selected_object_ids = val

    def should_be_displayed(self, attr: str, val: Any) -> bool:
        # Ignore the quantity value in the print statement when flag is True.
        if self._armour_attack_flag and attr == "quantity":
            return False
        if not self._armour_attack_flag and (attr == "armour_attack_quantity" or attr == "armour_attack_class"):
            return False

        return super().should_be_displayed(attr, val)

    def get_content_as_string(self, include_effect_definition=False) -> str:
        if self.effect_type not in effects.attributes:  # Unknown effect
            attributes_list = effects.empty_attributes
        else:
            attributes_list = effects.attributes[self.effect_type]

        return_string = ""
        for attribute in attributes_list:
            val = getattr(self, attribute)
            if not self.should_be_displayed(attribute, val):
                continue

            value_string = transform_effect_attr_value(self.effect_type, attribute, val, self._host_uuid)
            return_string += f"{attribute}: {value_string}\n"

        if return_string == "":
            return "<< No Attributes >>\n"

        if include_effect_definition:
            return f"{effects.effect_names[self.effect_type]}:\n{add_tabs(return_string, 1)}"
        return return_string

    def _update_armour_attack_flag(self):
        self._armour_attack_flag = _set_armour_attack_flag(self.effect_type, self.object_attributes)

    def _quantity_to_aa(self, quantity: int) -> Tuple[int, int]:
        """
        A function to convert the initial quantity value to the quantity and armor/attack values.
        Unfortunately this problem has to be solved in the object due to how specific this was implemented in DE.

        Args:
            quantity (int): the initial quantity value

        Returns:
            The one byte armor/attack class as int and one byte armor/attack quantity as int

        ----

        **Trigger Version 2.4**::

            Quantity value: (3, 5)
            00000000 00000000 00000011 000000101
                                aaq      aac

        Final 2/4 bytes are aaq (1 byte), and aac (1 byte). First 2 are unused. Max value of both is 255.

        **Trigger Version 2.5**::

            Quantity value: (3, 5)
            00000000 00000011 00000000 000000101
              aaq      aaq      aac      aac

        The 4/4 bytes are aaq (2 bytes), and aac (2 bytes). All are used. Max value of both is 65535.

        ----
        """
        trigger_version = getters.get_trigger_version(self._host_uuid)
        if trigger_version >= 2.5:
            return quantity >> 16, quantity & 65535
        return quantity >> 8, quantity & 255

    def _aa_to_quantity(self, aa_quantity: int, aa_class: int) -> int:
        """
        A function to convert the quantity and armor/attack field to a quantity value.
        Unfortunately this problem has to be solved in the object due to how specific this was implemented in DE.

        Args:
            aa_quantity (int): the armor quantity value
            aa_class (int): the armor/attack value

        Returns:
            The one byte quantity and one byte armor/attack value
        """
        trigger_version = getters.get_trigger_version(self._host_uuid)
        if trigger_version >= 2.5:
            return aa_class * 65536 + aa_quantity

        # Would use `aa_class << 8` - but apparently multiplication is faster
        return aa_class * 256 + aa_quantity

    def __str__(self):
        return f"[Effect] {self.get_content_as_string(include_effect_definition=True)}"


def _set_armour_attack_flag(effect_type, object_attributes) -> bool:
    return (effect_type in [EffectId.CHANGE_OBJECT_ATTACK, EffectId.CHANGE_OBJECT_ARMOR]) or \
           (effect_type == EffectId.MODIFY_ATTRIBUTE and object_attributes in [
               ObjectAttribute.ATTACK, ObjectAttribute.ARMOR
           ])



from enum import IntEnum


class GroupBy(IntEnum):
    """
    Enum used to select what way the triggers should be grouped when using the function 'copy_trigger_tree_per_player'.
    Each mode has it's own explanation explained in the corresponding docstring.
    """
    NONE = -1
    """
    Don't group triggers. Add triggers to the end of the list. This could mean the source triggers aren't near the 
    copies. All copies, however, will be grouped like the TRIGGER grouping mechanism - at the end of the list.
    """
    TRIGGER = 0
    """
    Group triggers by trigger. All copies of a trigger will be put directly below the 'source' trigger. 
    Example: If the 'source' trigger has display_index of 3, and another trigger in it's tree has 5. Assuming the 
    triggers were copied for all players, display indexes of the copies from the first trigger would be 4-10. The second
    trigger in the tree would be bumped up to 12 (from 5) and it's copies would use: 13-19. 
    
    Orders the given trigger on top with the child triggers after. Child triggers are ordered in the order they were
    found in (Depth-first).
    """
    PLAYER = 1
    """
    Group triggers by player. Warning: This mode could change some of the intended order of the triggers. All copies of
    the trigger will be put together sorted on player first, then trigger order. 
    Example: If the 'source' trigger has display_index of 3, and another trigger in it's tree has 5. Assuming the 
    triggers were copied for all players, display indexes of the copies from the firs trigger would be: 5, 7, 9, 11, 13,
    15 and 17. The second trigger would have moved to 4 (from 5) and it's copies would occupy slots: 6, 8, 10, 12, 14, 
    16 and 18. This way both PlayerId.ONE (assuming that's the source player) triggers will be together, same goes for 
    all other players. 
    
    In the example above can be seen how the second trigger was moved from 5 to 4. The unrelated trigger in between was
    moved to the end, 19. This might be an unwanted side effect, be aware.
    """



import sys

from AoE2ScenarioParser import settings
from AoE2ScenarioParser.helper.printers import warn

_block_below = (3, 8)
_notify_below = (3, 8)
_py_version = f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}"


def python_version_check():
    if sys.version_info < (3, 0):
        raise Exception("AoE2ScenarioParser does not support Python 2")

    if sys.version_info < _block_below:
        raise Exception(f"\n\nAoE2ScenarioParser requires the use of Python {'.'.join(map(str, _block_below))}.\n"
                        f"You are currently using version {_py_version}. You can download a newer version from: \n"
                        f"\thttps://www.python.org/downloads/\n")
    elif sys.version_info < _notify_below:
        if not settings.DISABLE_VERSION_WARNINGS:
            warn(
                f"AoE2ScenarioParser will update it's minimum python requirements to "
                f"Python {'.'.join(map(str, _notify_below))} in the (near) future.\n"
                f"You are currently using version {_py_version}. You can download a newer version from: \n"
                f"\thttps://www.python.org/downloads/\n\n"
                f"You can disable these warnings using:\n"
                "    from AoE2ScenarioParser import settings\n"
                "    settings.DISABLE_VERSION_WARNINGS = True"
            )



from __future__ import annotations

from collections import namedtuple
from typing import Type

Tile: Type[Tile] = namedtuple('Tile', ['x', 'y'])
"""NamedTuple for tiles. use tile.x or tile.y for coord access"""
